from hashlib import md5

from django.contrib.auth.models import AbstractBaseUser
from django.contrib.auth.models import AnonymousUser as DjangoAnonymousUser
from django.contrib.auth.models import PermissionsMixin
from django.contrib.auth.models import UserManager as BaseUserManager
from django.contrib.postgres.fields import ArrayField, HStoreField
from django.core.mail import send_mail
from django.db import models
from django.db.models import Q
from django.urls import reverse
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from ...acl.models import Role
from ...conf import settings
from ...core.utils import slugify
from ..avatars import store as avatars_store, delete_avatar
from ..signatures import is_user_signature_valid
from ..utils import hash_email
from .online import Online
from .rank import Rank


class UserManager(BaseUserManager):
    def _create_user(self, username, email, password, **extra_fields):
        """
        Create and save a user with the given username, email, and password.
        """
        if not username:
            raise ValueError("User must have an username.")
        if not email:
            raise ValueError("User must have an email address.")

        if not extra_fields.get("rank"):
            extra_fields["rank"] = Rank.objects.get_default()

        user = self.model(**extra_fields)
        user.set_username(username)
        user.set_email(email)
        user.set_password(password)

        now = extra_fields.get("joined_on", timezone.now())
        user.last_login = now
        user.joined_on = now

        user.save(using=self._db)
        self._assert_user_has_authenticated_role(user)

        Online.objects.create(user=user, last_click=now)

        return user

    def _assert_user_has_authenticated_role(self, user):
        authenticated_role = Role.objects.get(special_role="authenticated")
        if authenticated_role not in user.roles.all():
            user.roles.add(authenticated_role)
        user.update_acl_key()
        user.save(update_fields=["acl_key"])

    def create_user(self, username, email=None, password=None, **extra_fields):
        extra_fields.setdefault("is_staff", False)
        extra_fields.setdefault("is_superuser", False)
        return self._create_user(username, email, password, **extra_fields)

    def create_superuser(self, username, email, password=None, **extra_fields):
        extra_fields.setdefault("is_staff", True)
        extra_fields.setdefault("is_superuser", True)

        if extra_fields.get("is_staff") is not True:
            raise ValueError("Superuser must have is_staff=True.")
        if extra_fields.get("is_superuser") is not True:
            raise ValueError("Superuser must have is_superuser=True.")

        try:
            if not extra_fields.get("rank"):
                extra_fields["rank"] = Rank.objects.get(name=_("Forum team"))
        except Rank.DoesNotExist:
            pass

        return self._create_user(username, email, password, **extra_fields)

    def get_by_username(self, username):
        return self.get(slug=slugify(username))

    def get_by_email(self, email):
        return self.get(email_hash=hash_email(email))

    def get_by_username_or_email(self, login):
        if "@" in login:
            return self.get(email_hash=hash_email(login))
        return self.get(slug=slugify(login))


class User(AbstractBaseUser, PermissionsMixin):
    ACTIVATION_NONE = 0
    ACTIVATION_USER = 1
    ACTIVATION_ADMIN = 2

    SUBSCRIPTION_NONE = 0
    SUBSCRIPTION_NOTIFY = 1
    SUBSCRIPTION_ALL = 2

    SUBSCRIPTION_CHOICES = [
        (SUBSCRIPTION_NONE, _("No")),
        (SUBSCRIPTION_NOTIFY, _("Notify")),
        (SUBSCRIPTION_ALL, _("Notify with e-mail")),
    ]

    LIMIT_INVITES_TO_NONE = 0
    LIMIT_INVITES_TO_FOLLOWED = 1
    LIMIT_INVITES_TO_NOBODY = 2

    LIMIT_INVITES_TO_CHOICES = [
        (LIMIT_INVITES_TO_NONE, _("Everybody")),
        (LIMIT_INVITES_TO_FOLLOWED, _("Users I follow")),
        (LIMIT_INVITES_TO_NOBODY, _("Nobody")),
    ]

    # Note that "username" field is purely for shows.
    # When searching users by their names, always use lowercased string
    # and slug field instead that is normalized around DB engines
    # differences in case handling.
    username = models.CharField(max_length=30)
    slug = models.CharField(max_length=30, unique=True)

    # Misago stores user email in two fields:
    # "email" holds normalized email address
    # "email_hash" is lowercase hash of email address used to identify account
    # as well as enforcing on database level that no more than one user can be
    # using one email address
    email = models.EmailField(max_length=255, db_index=True)
    email_hash = models.CharField(max_length=32, unique=True)

    joined_on = models.DateTimeField(
        _("joined on"), default=timezone.now, db_index=True
    )
    joined_from_ip = models.GenericIPAddressField(null=True, blank=True)
    is_hiding_presence = models.BooleanField(default=False)

    rank = models.ForeignKey(
        "Rank", null=True, blank=True, on_delete=models.deletion.PROTECT
    )
    title = models.CharField(max_length=255, null=True, blank=True)
    requires_activation = models.PositiveIntegerField(default=ACTIVATION_NONE)

    is_staff = models.BooleanField(
        _("staff status"),
        default=False,
        help_text=_("Designates whether the user can log into admin sites."),
    )

    roles = models.ManyToManyField("misago_acl.Role")
    acl_key = models.CharField(max_length=12, null=True, blank=True)

    is_active = models.BooleanField(
        _("active"),
        db_index=True,
        default=True,
        help_text=_(
            "Designates whether this user should be treated as active. "
            "Unselect this instead of deleting accounts."
        ),
    )
    is_active_staff_message = models.TextField(null=True, blank=True)

    is_deleting_account = models.BooleanField(default=False)

    avatar_tmp = models.ImageField(
        max_length=255, upload_to=avatars_store.upload_to, null=True, blank=True
    )
    avatar_src = models.ImageField(
        max_length=255, upload_to=avatars_store.upload_to, null=True, blank=True
    )
    avatar_crop = models.CharField(max_length=255, null=True, blank=True)
    avatars = models.JSONField(null=True, blank=True)
    is_avatar_locked = models.BooleanField(default=False)
    avatar_lock_user_message = models.TextField(null=True, blank=True)
    avatar_lock_staff_message = models.TextField(null=True, blank=True)

    signature = models.TextField(null=True, blank=True)
    signature_parsed = models.TextField(null=True, blank=True)
    signature_checksum = models.CharField(max_length=64, null=True, blank=True)
    is_signature_locked = models.BooleanField(default=False)
    signature_lock_user_message = models.TextField(null=True, blank=True)
    signature_lock_staff_message = models.TextField(null=True, blank=True)

    followers = models.PositiveIntegerField(default=0)
    following = models.PositiveIntegerField(default=0)

    follows = models.ManyToManyField(
        "self", related_name="followed_by", symmetrical=False
    )
    blocks = models.ManyToManyField(
        "self", related_name="blocked_by", symmetrical=False
    )

    limits_private_thread_invites_to = models.PositiveIntegerField(
        default=LIMIT_INVITES_TO_NONE, choices=LIMIT_INVITES_TO_CHOICES
    )
    unread_private_threads = models.PositiveIntegerField(default=0)
    sync_unread_private_threads = models.BooleanField(default=False)

    subscribe_to_started_threads = models.PositiveIntegerField(
        default=SUBSCRIPTION_NONE, choices=SUBSCRIPTION_CHOICES
    )
    subscribe_to_replied_threads = models.PositiveIntegerField(
        default=SUBSCRIPTION_NONE, choices=SUBSCRIPTION_CHOICES
    )

    threads = models.PositiveIntegerField(default=0)
    posts = models.PositiveIntegerField(default=0, db_index=True)

    last_posted_on = models.DateTimeField(null=True, blank=True)

    profile_fields = HStoreField(default=dict)
    agreements = ArrayField(models.PositiveIntegerField(), default=list)

    USERNAME_FIELD = "slug"
    REQUIRED_FIELDS = ["email"]

    objects = UserManager()

    class Meta:
        indexes = [
            models.Index(
                name="misago_user_is_staff_part",
                fields=["is_staff"],
                condition=Q(is_staff=True),
            ),
            models.Index(
                name="misago_user_requires_acti_part",
                fields=["requires_activation"],
                condition=Q(requires_activation__gt=0),
            ),
            models.Index(
                name="misago_user_is_deleting_a_part",
                fields=["is_deleting_account"],
                condition=Q(is_deleting_account=True),
            ),
        ]

    def clean(self):
        self.username = self.normalize_username(self.username)
        self.email = UserManager.normalize_email(self.email)

    def lock(self):
        """locks user in DB, shortcut for locking user model in views"""
        return User.objects.select_for_update().get(pk=self.pk)

    def delete(self, *args, **kwargs):
        if kwargs.pop("delete_content", False):
            self.delete_content()

        username = kwargs.pop("anonymous_username", None)
        if username:
            self.anonymize_data(username)
        else:
            raise ValueError("user.delete() requires 'anonymous_username' argument")

        delete_avatar(self)

        return super().delete(*args, **kwargs)

    def delete_content(self):
        from ..signals import delete_user_content

        delete_user_content.send(sender=self)

    def mark_for_delete(self):
        self.is_active = False
        self.is_deleting_account = True
        self.save(update_fields=["is_active", "is_deleting_account"])

    def anonymize_data(self, anonymous_username):
        """Replaces username with anonymized one, then send anonymization signal.

        Items associated with this user then anonymize their user-specific data
        like username or IP addresses.
        """
        self.username = anonymous_username
        self.slug = slugify(self.username)

        from ..signals import anonymize_user_data

        anonymize_user_data.send(sender=self)

    @property
    def requires_activation_by_admin(self):
        return self.requires_activation == self.ACTIVATION_ADMIN

    @property
    def requires_activation_by_user(self):
        return self.requires_activation == self.ACTIVATION_USER

    @property
    def can_be_messaged_by_everyone(self):
        preference = self.limits_private_thread_invites_to
        return preference == self.LIMIT_INVITES_TO_NONE

    @property
    def can_be_messaged_by_followed(self):
        preference = self.limits_private_thread_invites_to
        return preference == self.LIMIT_INVITES_TO_FOLLOWED

    @property
    def can_be_messaged_by_nobody(self):
        preference = self.limits_private_thread_invites_to
        return preference == self.LIMIT_INVITES_TO_NOBODY

    @property
    def has_valid_signature(self):
        return is_user_signature_valid(self)

    def get_absolute_url(self):
        return reverse("misago:user", kwargs={"slug": self.slug, "pk": self.pk})

    def get_username(self):
        """dirty hack: return real username instead of normalized slug"""
        return self.username

    def get_full_name(self):
        return self.username

    def get_short_name(self):
        return self.username

    def get_real_name(self):
        # Bug: https://github.com/rafalp/Misago/issues/1352
        # On some very rare cases self.profile_fields is deserialized as string
        # by Django's ORM. I am unable to reproduce this, but in case when this
        # bug occurs, this code branch will print extra information about it
        if not isinstance(self.profile_fields, dict):
            from django.db import connections
            from django.db.backends.signals import connection_created
            from django.contrib.postgres.signals import get_hstore_oids

            receivers = ", ".join([str(r[1]()) for r in connection_created.receivers])
            dead_receivers = "TRUE" if connection_created._dead_receivers else "FALSE"

            valid_oids = None
            cached_oids = get_hstore_oids("default")

            with connections["default"].cursor() as cursor:
                cursor.execute(
                    "SELECT t.oid, typarray "
                    "FROM pg_type t "
                    "JOIN pg_namespace ns ON typnamespace = ns.oid "
                    "WHERE typname = 'hstore'"
                )
                oids = []
                array_oids = []
                for row in cursor:
                    oids.append(row[0])
                    array_oids.append(row[1])
                valid_oids = tuple(oids), tuple(array_oids)

            raise RuntimeError(
                f"'profile_fields' has wrong type! Please post this WHOLE message on https://github.com/rafalp/Misago/issues/1352 "
                f"OID: '{cached_oids}' (valid: '{valid_oids}') "
                f"Receivers: '{receivers}' (has dead: {dead_receivers}) "
                f"Repr: {repr(self.profile_fields)}"
            )

        return self.profile_fields.get("real_name")

    def set_username(self, new_username, changed_by=None):
        new_username = self.normalize_username(new_username)
        if new_username != self.username:
            old_username = self.username
            self.username = new_username
            self.slug = slugify(new_username)

            if self.pk:
                changed_by = changed_by or self
                namechange = self.record_name_change(
                    changed_by, new_username, old_username
                )

                from ..signals import username_changed

                username_changed.send(sender=self)

                return namechange

    def record_name_change(self, changed_by, new_username, old_username):
        return self.namechanges.create(
            new_username=new_username,
            old_username=old_username,
            changed_by=changed_by,
            changed_by_username=changed_by.username,
        )

    def set_email(self, new_email):
        self.email = UserManager.normalize_email(new_email)
        self.email_hash = hash_email(new_email)

    def get_any_title(self):
        return self.title or self.rank.title or self.rank.name

    def get_roles(self):
        roles_pks = []
        roles_dict = {}

        for role in self.roles.all():
            roles_pks.append(role.pk)
            role.origin = self
            roles_dict[role.pk] = role

        if self.rank:
            for role in self.rank.roles.all():
                if role.pk not in roles_pks:
                    role.origin = self.rank
                    roles_pks.append(role.pk)
                    roles_dict[role.pk] = role

        return [roles_dict[r] for r in sorted(roles_pks)]

    def update_acl_key(self):
        roles_pks = []
        for role in self.get_roles():
            if role.origin == "self":
                roles_pks.append("u%s" % role.pk)
            else:
                roles_pks.append("%s:%s" % (self.rank.pk, role.pk))

        self.acl_key = md5(",".join(roles_pks).encode()).hexdigest()[:12]

    def email_user(self, subject, message, from_email=None, **kwargs):
        """sends an email to this user (for compat with Django)"""
        send_mail(subject, message, from_email, [self.email], **kwargs)

    def is_following(self, user_or_id):
        try:
            user_id = user_or_id.id
        except AttributeError:
            user_id = user_or_id

        try:
            self.follows.get(id=user_id)
            return True
        except User.DoesNotExist:
            return False

    def is_blocking(self, user_or_id):
        try:
            user_id = user_or_id.id
        except AttributeError:
            user_id = user_or_id

        try:
            self.blocks.get(id=user_id)
            return True
        except User.DoesNotExist:
            return False


class UsernameChange(models.Model):
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, related_name="namechanges", on_delete=models.CASCADE
    )
    changed_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        null=True,
        blank=True,
        related_name="user_renames",
        on_delete=models.SET_NULL,
    )
    changed_by_username = models.CharField(max_length=30)
    changed_on = models.DateTimeField(default=timezone.now)
    new_username = models.CharField(max_length=255)
    old_username = models.CharField(max_length=255)

    class Meta:
        get_latest_by = "changed_on"

    def set_change_author(self, user):
        self.changed_by = user
        self.changed_by_username = user.username


class AnonymousUser(DjangoAnonymousUser):
    acl_key = "anonymous"

    @property
    def acl_cache(self):
        raise Exception("AnonymousUser.acl_cache has been removed")

    @acl_cache.setter
    def acl_cache(self, value):
        raise TypeError("AnonymousUser instances can't be made ACL aware")

    def get_roles(self):
        try:
            return [Role.objects.get(special_role="anonymous")]
        except Role.DoesNotExist:
            raise RuntimeError("Anonymous user role not found.")

    def update_acl_key(self):
        raise TypeError("Can't update ACL key on anonymous users")



from .rank import Rank
from .online import Online
from .user import AnonymousUser, User, UsernameChange
from .activityranking import ActivityRanking
from .avatar import Avatar
from .audittrail import AuditTrail
from .avatargallery import AvatarGallery
from .ban import Ban, BanCache
from .datadownload import DataDownload
from .deleteduser import DeletedUser



from .models import CacheVersion
from .utils import generate_version_string


def get_cache_versions():
    queryset = CacheVersion.objects.all()
    return {i.cache: i.version for i in queryset}


def invalidate_cache(cache_name):
    CacheVersion.objects.filter(cache=cache_name).update(
        version=generate_version_string()
    )


def invalidate_all_caches():
    for cache_name in get_cache_versions().keys():
        CacheVersion.objects.filter(cache=cache_name).update(
            version=generate_version_string()
        )



import pytest
from django.urls import reverse

from ....admin.test import AdminTestCase
from ....test import assert_contains
from ...models import Agreement


@pytest.fixture
def list_url(admin_client):
    response = admin_client.get(reverse("misago:admin:settings:agreements:index"))
    return response["location"]


@pytest.fixture
def agreement(superuser):
    return Agreement.objects.create(
        type=Agreement.TYPE_TOS,
        title="Test TOS",
        link="https://rules.com",
        created_by=superuser,
        created_by_name=superuser.username,
    )


@pytest.fixture
def active_agreement(agreement):
    agreement.is_active = True
    agreement.save()
    return agreement


@pytest.fixture
def other_agreement(superuser):
    return Agreement.objects.create(
        type=Agreement.TYPE_TOS,
        title="Test TOS",
        link="https://rules.com",
        created_by=superuser,
        created_by_name=superuser.username,
    )


def test_nav_contains_agreements_link(admin_client, list_url):
    response = admin_client.get(list_url)
    assert_contains(response, reverse("misago:admin:settings:agreements:index"))


def test_empty_list_renders(admin_client, list_url):
    response = admin_client.get(list_url)
    assert response.status_code == 200


def test_list_renders_agreement(admin_client, list_url, agreement):
    response = admin_client.get(list_url)
    assert_contains(response, agreement.title)


def test_agreements_can_be_mass_deleted(admin_client, list_url, superuser):
    agreements = []
    for _ in range(10):
        agreement = Agreement.objects.create(
            type=Agreement.TYPE_TOS,
            title="Test TOS",
            link="https://rules.com",
            created_by=superuser,
            created_by_name=superuser.username,
        )
        agreements.append(agreement.pk)

    response = admin_client.post(
        list_url, data={"action": "delete", "selected_items": agreements}
    )
    assert response.status_code == 302
    assert Agreement.objects.count() == 0


def test_creation_form_renders(admin_client):
    response = admin_client.get(reverse("misago:admin:settings:agreements:new"))
    assert response.status_code == 200


def test_form_creates_new_agreement(admin_client):
    response = admin_client.post(
        reverse("misago:admin:settings:agreements:new"),
        {
            "type": Agreement.TYPE_TOS,
            "title": "Test TOS",
            "text": "Lorem ipsum dolor met sit amet elit",
            "link": "https://example.com/rules/",
        },
    )

    agreement = Agreement.objects.get()
    assert agreement.type == Agreement.TYPE_TOS
    assert agreement.title == "Test TOS"
    assert agreement.text == "Lorem ipsum dolor met sit amet elit"
    assert agreement.link == "https://example.com/rules/"


def test_form_sets_new_agreement_creator(admin_client, superuser):
    response = admin_client.post(
        reverse("misago:admin:settings:agreements:new"),
        {
            "type": Agreement.TYPE_TOS,
            "title": "Test TOS",
            "text": "Lorem ipsum dolor met sit amet elit",
            "link": "https://example.com/rules/",
        },
    )

    agreement = Agreement.objects.get()
    assert agreement.created_by == superuser
    assert agreement.created_by_name == superuser.username


def test_form_creates_active_agreement(mocker, admin_client):
    set_agreement_as_active = mocker.patch(
        "misago.legal.admin.forms.set_agreement_as_active"
    )
    response = admin_client.post(
        reverse("misago:admin:settings:agreements:new"),
        {
            "type": Agreement.TYPE_TOS,
            "is_active": "1",
            "title": "Test TOS",
            "text": "Lorem ipsum dolor met sit amet elit",
            "link": "https://example.com/rules/",
        },
    )

    agreement = Agreement.objects.get()
    assert agreement.is_active
    set_agreement_as_active.assert_called_once_with(agreement)


def test_newly_created_active_agreement_replaces_current_one(
    admin_client, active_agreement
):
    response = admin_client.post(
        reverse("misago:admin:settings:agreements:new"),
        {
            "type": Agreement.TYPE_TOS,
            "is_active": "1",
            "title": "Test TOS",
            "text": "Lorem ipsum dolor met sit amet elit",
            "link": "https://example.com/rules/",
        },
    )

    agreement = Agreement.objects.exclude(pk=active_agreement.pk).get()
    assert agreement.is_active

    active_agreement.refresh_from_db()
    assert not active_agreement.is_active


def test_edit_form_renders(admin_client, agreement):
    response = admin_client.get(
        reverse("misago:admin:settings:agreements:edit", kwargs={"pk": agreement.pk})
    )
    assert_contains(response, agreement.title)


def test_edit_form_updates_agreement(admin_client, agreement):
    response = admin_client.post(
        reverse("misago:admin:settings:agreements:edit", kwargs={"pk": agreement.pk}),
        data={
            "type": Agreement.TYPE_TOS,
            "title": "Test Edited",
            "text": "Sit amet elit",
            "link": "https://example.com/terms/",
        },
    )
    assert response.status_code == 302

    agreement.refresh_from_db()
    assert agreement.type == Agreement.TYPE_TOS
    assert agreement.title == "Test Edited"
    assert agreement.text == "Sit amet elit"
    assert agreement.link == "https://example.com/terms/"


def test_edit_form_updates_agreement_modified_entry(admin_client, agreement, superuser):
    response = admin_client.post(
        reverse("misago:admin:settings:agreements:edit", kwargs={"pk": agreement.pk}),
        data={
            "type": Agreement.TYPE_TOS,
            "title": "Test Edited",
            "text": "Sit amet elit",
            "link": "https://example.com/terms/",
        },
    )
    assert response.status_code == 302

    agreement.refresh_from_db()
    assert agreement.last_modified_on
    assert agreement.last_modified_by == superuser
    assert agreement.last_modified_by_name == superuser.username


def test_edit_form_changes_active_agreement(
    admin_client, active_agreement, other_agreement
):
    response = admin_client.post(
        reverse(
            "misago:admin:settings:agreements:edit", kwargs={"pk": other_agreement.pk}
        ),
        data={
            "type": Agreement.TYPE_TOS,
            "is_active": "1",
            "title": "Test Edited",
            "text": "Sit amet elit",
            "link": "https://example.com/terms/",
        },
    )
    assert response.status_code == 302

    active_agreement.refresh_from_db()
    assert not active_agreement.is_active

    other_agreement.refresh_from_db()
    assert other_agreement.is_active


def test_edit_form_disables_active_agreement(admin_client, active_agreement):
    response = admin_client.post(
        reverse(
            "misago:admin:settings:agreements:edit", kwargs={"pk": active_agreement.pk}
        ),
        data={
            "type": Agreement.TYPE_TOS,
            "is_active": "0",
            "title": "Test Edited",
            "text": "Sit amet elit",
            "link": "https://example.com/terms/",
        },
    )
    assert response.status_code == 302

    active_agreement.refresh_from_db()
    assert not active_agreement.is_active


def test_agreement_can_be_deleted(admin_client, agreement):
    response = admin_client.post(
        reverse("misago:admin:settings:agreements:delete", kwargs={"pk": agreement.pk})
    )
    assert response.status_code == 302

    with pytest.raises(Agreement.DoesNotExist):
        agreement.refresh_from_db()


def test_active_agreement_can_be_deleted(admin_client, active_agreement):
    response = admin_client.post(
        reverse(
            "misago:admin:settings:agreements:delete",
            kwargs={"pk": active_agreement.pk},
        )
    )
    assert response.status_code == 302

    with pytest.raises(Agreement.DoesNotExist):
        active_agreement.refresh_from_db()


def test_agreement_can_be_set_as_active(admin_client, agreement):
    response = admin_client.post(
        reverse(
            "misago:admin:settings:agreements:set-as-active",
            kwargs={"pk": agreement.pk},
        )
    )
    assert response.status_code == 302

    agreement.refresh_from_db()
    assert agreement.is_active


def test_active_agreement_can_be_changed(
    admin_client, active_agreement, other_agreement
):
    response = admin_client.post(
        reverse(
            "misago:admin:settings:agreements:set-as-active",
            kwargs={"pk": other_agreement.pk},
        )
    )
    assert response.status_code == 302

    active_agreement.refresh_from_db()
    assert not active_agreement.is_active

    other_agreement.refresh_from_db()
    assert other_agreement.is_active


def test_active_agreement_can_be_disabled(admin_client, active_agreement):
    response = admin_client.post(
        reverse(
            "misago:admin:settings:agreements:disable",
            kwargs={"pk": active_agreement.pk},
        )
    )
    assert response.status_code == 302

    active_agreement.refresh_from_db()
    assert not active_agreement.is_active



from django.contrib import messages
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from ....admin.views import generic
from ...models import Rank
from ..forms import RankForm


class RankAdmin(generic.AdminBaseMixin):
    root_link = "misago:admin:ranks:index"
    model = Rank
    form_class = RankForm
    templates_dir = "misago/admin/ranks"
    message_404 = _("Requested rank does not exist.")

    def update_roles(self, target, roles):
        target.roles.clear()
        if roles:
            target.roles.add(*roles)

    def handle_form(self, form, request, target):
        super().handle_form(form, request, target)
        self.update_roles(target, form.cleaned_data["roles"])


class RanksList(RankAdmin, generic.ListView):
    ordering = (("order", None),)


class NewRank(RankAdmin, generic.ModelFormView):
    message_submit = _('New rank "%(name)s" has been saved.')


class EditRank(RankAdmin, generic.ModelFormView):
    message_submit = _('Rank "%(name)s" has been edited.')


class DeleteRank(RankAdmin, generic.ButtonView):
    def check_permissions(self, request, target):
        message_format = {"name": target.name}
        if target.is_default:
            message = _('Rank "%(name)s" is default rank and can\'t be deleted.')
            return message % message_format
        if target.user_set.exists():
            message = _('Rank "%(name)s" is assigned to users and can\'t be deleted.')
            return message % message_format

    def button_action(self, request, target):
        target.delete()
        message = _('Rank "%(name)s" has been deleted.')
        messages.success(request, message % {"name": target.name})


class MoveDownRank(RankAdmin, generic.ButtonView):
    def button_action(self, request, target):
        try:
            other_target = Rank.objects.filter(order__gt=target.order)
            other_target = other_target.earliest("order")
        except Rank.DoesNotExist:
            other_target = None

        if other_target:
            other_target.order, target.order = target.order, other_target.order
            other_target.save(update_fields=["order"])
            target.save(update_fields=["order"])

            message = _('Rank "%(name)s" has been moved below "%(other)s".')
            targets_names = {"name": target.name, "other": other_target.name}
            messages.success(request, message % targets_names)


class MoveUpRank(RankAdmin, generic.ButtonView):
    def button_action(self, request, target):
        try:
            other_target = Rank.objects.filter(order__lt=target.order)
            other_target = other_target.latest("order")
        except Rank.DoesNotExist:
            other_target = None

        if other_target:
            other_target.order, target.order = target.order, other_target.order
            other_target.save(update_fields=["order"])
            target.save(update_fields=["order"])

            message = _('Rank "%(name)s" has been moved above "%(other)s".')
            targets_names = {"name": target.name, "other": other_target.name}
            messages.success(request, message % targets_names)


class RankUsers(RankAdmin, generic.TargetedView):
    def real_dispatch(self, request, target):
        redirect_url = reverse("misago:admin:users:index")
        return redirect("%s?rank=%s" % (redirect_url, target.pk))


class DefaultRank(RankAdmin, generic.ButtonView):
    def check_permissions(self, request, target):
        if target.is_default:
            message = _('Rank "%(name)s" is already default.')
            return message % {"name": target.name}

    def button_action(self, request, target):
        Rank.objects.make_rank_default(target)
        message = _('Rank "%(name)s" has been made default.')
        messages.success(request, message % {"name": target.name})



from django.shortcuts import redirect
from django.utils.deprecation import MiddlewareMixin

from .auth import (
    is_admin_authorized,
    remove_admin_authorization,
    update_admin_authorization,
)
from .views import get_protected_namespace
from .views.auth import login


class AdminAuthMiddleware(MiddlewareMixin):
    def process_view(self, request, view_func, view_args, view_kwargs):
        request.admin_namespace = get_protected_namespace(request)
        if request.admin_namespace:
            return self.check_admin_authorization(request)

    def check_admin_authorization(self, request):
        if not is_admin_authorized(request):
            remove_admin_authorization(request)
            if request.resolver_match.url_name == "index":
                return login(request)
            return redirect("%s:index" % request.admin_namespace)

        update_admin_authorization(request)



from io import StringIO

from django.core.management import call_command

from ..management.commands import populateonlinetracker
from ..models import Online


def test_management_command_creates_online_tracker_for_user_without_one(user):
    Online.objects.filter(user=user).delete()
    assert not Online.objects.filter(user=user).exists()

    call_command(populateonlinetracker.Command(), stdout=StringIO())
    assert Online.objects.filter(user=user).exists()


def test_management_command_displays_message_with_number_of_created_trackers(user):
    Online.objects.filter(user=user).delete()
    assert not Online.objects.filter(user=user).exists()

    out = StringIO()
    call_command(populateonlinetracker.Command(), stdout=out)
    command_output = out.getvalue().splitlines()[0].strip()
    assert command_output == "Tracker entries created: 1"



from django.http import HttpResponse

from ..middleware import FrontendContextMiddleware


class MockRequest:
    pass


def get_response(*_):
    return HttpResponse("OK!")


def test_frontend_middleware_sets_frontend_context_dict_on_request():
    request = MockRequest()
    FrontendContextMiddleware(get_response)(request)
    assert request.frontend_context == {}



from datetime import timedelta

import pytest
from django.utils import timezone

from ...conf.test import override_dynamic_settings
from ..poststracker import make_read_aware, save_read


def test_falsy_value_can_be_made_read_aware(request_mock):
    make_read_aware(request_mock, None)
    make_read_aware(request_mock, False)


def test_empty_list_can_be_made_read_aware(request_mock):
    make_read_aware(request_mock, [])


@pytest.fixture
def read_post(user, post):
    save_read(user, post)
    return post


def test_tracked_post_is_marked_as_not_read_and_new(request_mock, post):
    make_read_aware(request_mock, post)
    assert not post.is_read
    assert post.is_new


@override_dynamic_settings(readtracker_cutoff=3)
def test_not_tracked_post_is_marked_as_read_and_not_new(request_mock, post):
    post.posted_on = timezone.now() - timedelta(days=4)
    post.save()

    make_read_aware(request_mock, post)
    assert post.is_read
    assert not post.is_new


def test_tracked_read_post_is_marked_as_read_and_not_new(request_mock, read_post):
    make_read_aware(request_mock, read_post)
    assert read_post.is_read
    assert not read_post.is_new


@override_dynamic_settings(readtracker_cutoff=3)
def test_not_tracked_read_post_is_marked_as_read_and_not_new(request_mock, read_post):
    read_post.posted_on = timezone.now() - timedelta(days=4)
    read_post.save()

    make_read_aware(request_mock, read_post)
    assert read_post.is_read
    assert not read_post.is_new


def test_iterable_of_posts_can_be_made_read_aware(request_mock, post):
    make_read_aware(request_mock, [post])
    assert not post.is_read
    assert post.is_new


def test_tracked_post_read_by_other_user_is_marked_as_not_read_and_new(
    request_mock, other_user, post
):
    save_read(other_user, post)
    make_read_aware(request_mock, post)
    assert not post.is_read
    assert post.is_new


def test_tracked_post_is_marked_as_read_and_not_new_for_anonymous_user(
    anonymous_request_mock, post
):
    make_read_aware(anonymous_request_mock, post)
    assert post.is_read
    assert not post.is_new


@override_dynamic_settings(readtracker_cutoff=3)
def test_not_tracked_post_is_marked_as_read_and_not_new_for_anonymous_user(
    anonymous_request_mock, post
):
    post.posted_on = timezone.now() - timedelta(days=4)
    post.save()

    make_read_aware(anonymous_request_mock, post)
    assert post.is_read
    assert not post.is_new



from functools import partial, reduce


def apply_filters(action, hook, *args, **kwargs):
    if not hook:
        return action(*args, **kwargs)

    filtered_action = reduce(
        lambda chained_fns, next_fn: partial(next_fn, chained_fns), hook, action
    )
    return filtered_action(*args, **kwargs)



from .. import test
from ...users.test import create_test_user
from ..models import ThreadParticipant
from .test_privatethreads import PrivateThreadsTestCase


class PrivateThreadReplyApiTestCase(PrivateThreadsTestCase):
    def setUp(self):
        super().setUp()

        self.thread = test.post_thread(self.category, poster=self.user)
        self.api_link = self.thread.get_posts_api_url()

        self.other_user = create_test_user("OtherUser", "otheruser@example.com")

    def test_reply_private_thread(self):
        """api sets other private thread participants sync thread flag"""
        ThreadParticipant.objects.set_owner(self.thread, self.user)
        ThreadParticipant.objects.add_participants(self.thread, [self.other_user])

        response = self.client.post(
            self.api_link, data={"post": "This is test response!"}
        )
        self.assertEqual(response.status_code, 200)

        # don't count private thread replies
        self.reload_user()
        self.assertEqual(self.user.threads, 0)
        self.assertEqual(self.user.posts, 0)

        self.assertEqual(self.user.audittrail_set.count(), 1)

        # valid user was flagged to sync
        self.user.refresh_from_db()
        self.assertFalse(self.user.sync_unread_private_threads)

        self.other_user.refresh_from_db()
        self.assertTrue(self.other_user.sync_unread_private_threads)



import re
from xml.etree.ElementTree import SubElement

import markdown
from django.utils.crypto import get_random_string
from markdown.blockprocessors import BlockProcessor
from markdown.preprocessors import Preprocessor

QUOTE_START = get_random_string(32)
QUOTE_END = get_random_string(32)


class QuoteExtension(markdown.Extension):
    def extendMarkdown(self, md):
        md.registerExtension(self)

        md.preprocessors.register(QuotePreprocessor(md), "misago_bbcode_quote", 200)
        md.parser.blockprocessors.register(
            QuoteBlockProcessor(md.parser), "misago_bbcode_quote", 90
        )


class QuotePreprocessor(Preprocessor):
    QUOTE_BLOCK_RE = re.compile(
        r"""
\[quote\](?P<text>.*?)\[/quote\]
""".strip(),
        re.IGNORECASE | re.MULTILINE | re.DOTALL,
    )
    QUOTE_BLOCK_TITLE_RE = re.compile(
        r"""
\[quote=("?)(?P<title>.*?)("?)](?P<text>.*?)\[/quote\]
""".strip(),
        re.IGNORECASE | re.MULTILINE | re.DOTALL,
    )

    def run(self, lines):
        text = "\n".join(lines)
        while self.QUOTE_BLOCK_RE.search(text):
            text = self.QUOTE_BLOCK_RE.sub(self.replace, text)
        while self.QUOTE_BLOCK_TITLE_RE.search(text):
            text = self.QUOTE_BLOCK_TITLE_RE.sub(self.replace_titled, text)
        return text.split("\n")

    def replace(self, matchobj):
        text = matchobj.group("text")
        return "\n\n%s\n\n%s\n\n%s\n\n" % (QUOTE_START, text, QUOTE_END)

    def replace_titled(self, matchobj):
        title = matchobj.group("title").strip()
        text = matchobj.group("text")

        if title:
            return "\n\n%s%s\n\n%s\n\n%s\n\n" % (QUOTE_START, title, text, QUOTE_END)
        return "\n\n%s\n\n%s\n\n%s\n\n" % (QUOTE_START, text, QUOTE_END)


class QuoteBlockProcessor(BlockProcessor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._title = None
        self._quote = 0
        self._children = []

    def test(self, parent, block):
        return block.strip().startswith(QUOTE_START) or self._quote

    def run(self, parent, blocks):
        block = blocks.pop(0)
        if block.strip().startswith(QUOTE_START):
            self._quote += 1
            if self._quote == 1:
                self._title = block[len(QUOTE_START) :].strip() or None

        self._children.append(block)

        if block.strip() == QUOTE_END:
            self._quote -= 1

        if not self._quote:
            children, self._children = self._children[1:-1], []
            title, self._title = self._title, None

            aside = SubElement(parent, "aside")
            aside.set("class", "quote-block")

            heading = SubElement(aside, "div")
            heading.set("class", "quote-heading")
            heading.set("data-noquote", "1")

            blockquote = SubElement(aside, "blockquote")
            blockquote.set("class", "quote-body")
            blockquote.set("data-block", "quote")

            if title:
                blockquote.set("data-author", title)

            if title:
                heading.text = title

            self.parser.parseBlocks(blockquote, children)



import json

from django.urls import reverse

from .. import test
from ...acl import useracl
from ...acl.objectacl import add_acl_to_obj
from ...categories.models import Category
from ...conf.test import override_dynamic_settings
from ...conftest import get_cache_versions
from ...readtracker import poststracker
from ..models import Poll, PollVote, Post, Thread
from ..serializers import ThreadsListSerializer
from ..test import patch_category_acl, patch_other_category_acl
from .test_threads_api import ThreadsApiTestCase

cache_versions = get_cache_versions()


class ThreadsMergeApiTests(ThreadsApiTestCase):
    def setUp(self):
        super().setUp()
        self.api_link = reverse("misago:api:thread-merge")

        Category(name="Other Category", slug="other-category").insert_at(
            self.category, position="last-child", save=True
        )
        self.other_category = Category.objects.get(slug="other-category")

    def test_merge_no_threads(self):
        """api validates if we are trying to merge no threads"""
        response = self.client.post(self.api_link, content_type="application/json")
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "You have to select at least two threads to merge."},
        )

    def test_merge_empty_threads(self):
        """api validates if we are trying to empty threads list"""
        response = self.client.post(
            self.api_link, json.dumps({"threads": []}), content_type="application/json"
        )
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "You have to select at least two threads to merge."},
        )

    def test_merge_invalid_threads(self):
        """api validates if we are trying to merge invalid thread ids"""
        response = self.client.post(
            self.api_link,
            json.dumps({"threads": "abcd"}),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": 'Expected a list of items but got type "str".'}
        )

        response = self.client.post(
            self.api_link,
            json.dumps({"threads": ["a", "-", "c"]}),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": ["One or more thread ids received were invalid."]},
        )

    def test_merge_single_thread(self):
        """api validates if we are trying to merge single thread"""
        response = self.client.post(
            self.api_link,
            json.dumps({"threads": [self.thread.id]}),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "You have to select at least two threads to merge."},
        )

    def test_merge_with_nonexisting_thread(self):
        """api validates if we are trying to merge with invalid thread"""
        response = self.client.post(
            self.api_link,
            json.dumps({"threads": [self.thread.id, self.thread.id + 1000]}),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "One or more threads to merge could not be found."},
        )

    def test_merge_with_invisible_thread(self):
        """api validates if we are trying to merge with inaccesible thread"""
        unaccesible_thread = test.post_thread(category=self.other_category)

        response = self.client.post(
            self.api_link,
            json.dumps({"threads": [self.thread.id, unaccesible_thread.id]}),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "One or more threads to merge could not be found."},
        )

    def test_merge_no_permission(self):
        """api validates permission to merge threads"""
        thread = test.post_thread(category=self.category)

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "category": self.category.id,
                    "title": "Lorem ipsum dolor",
                    "threads": [self.thread.id, thread.id],
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            [
                {
                    "id": thread.pk,
                    "title": thread.title,
                    "errors": ["You can't merge threads in this category."],
                },
                {
                    "id": self.thread.pk,
                    "title": self.thread.title,
                    "errors": ["You can't merge threads in this category."],
                },
            ],
        )

    @patch_other_category_acl()
    @patch_category_acl({"can_merge_threads": True, "can_close_threads": False})
    def test_thread_category_is_closed(self):
        """api validates if thread's category is open"""
        other_thread = test.post_thread(self.category)

        self.category.is_closed = True
        self.category.save()

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "category": self.other_category.id,
                    "title": "Lorem ipsum dolor",
                    "threads": [self.thread.id, other_thread.id],
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            [
                {
                    "id": other_thread.id,
                    "title": other_thread.title,
                    "errors": [
                        "This category is closed. You can't merge it's threads."
                    ],
                },
                {
                    "id": self.thread.id,
                    "title": self.thread.title,
                    "errors": [
                        "This category is closed. You can't merge it's threads."
                    ],
                },
            ],
        )

    @patch_other_category_acl()
    @patch_category_acl({"can_merge_threads": True, "can_close_threads": False})
    def test_thread_is_closed(self):
        """api validates if thread is open"""
        other_thread = test.post_thread(self.category)

        other_thread.is_closed = True
        other_thread.save()

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "category": self.other_category.id,
                    "title": "Lorem ipsum dolor",
                    "threads": [self.thread.id, other_thread.id],
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            [
                {
                    "id": other_thread.id,
                    "title": other_thread.title,
                    "errors": [
                        "This thread is closed. You can't merge it with other threads."
                    ],
                }
            ],
        )

    @override_dynamic_settings(threads_per_page=4)
    @patch_category_acl({"can_merge_threads": True})
    def test_merge_too_many_threads(self):
        """api rejects too many threads to merge"""
        threads = []
        for _ in range(5):
            threads.append(test.post_thread(category=self.category).pk)

        response = self.client.post(
            self.api_link,
            json.dumps({"threads": threads}),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "No more than 4 threads can be merged at a single time."},
        )

    @patch_category_acl({"can_merge_threads": True})
    def test_merge_no_final_thread(self):
        """api rejects merge because no data to merge threads was specified"""
        thread = test.post_thread(category=self.category)

        response = self.client.post(
            self.api_link,
            json.dumps({"threads": [self.thread.id, thread.id]}),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "title": ["This field is required."],
                "category": ["This field is required."],
            },
        )

    @patch_category_acl({"can_merge_threads": True})
    def test_merge_invalid_final_title(self):
        """api rejects merge because final thread title was invalid"""
        thread = test.post_thread(category=self.category)

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "threads": [self.thread.id, thread.id],
                    "title": "$$$",
                    "category": self.category.id,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "title": [
                    "Thread title should be at least 5 characters long (it has 3)."
                ]
            },
        )

    @patch_category_acl({"can_merge_threads": True})
    def test_merge_invalid_category(self):
        """api rejects merge because final category was invalid"""
        thread = test.post_thread(category=self.category)

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "threads": [self.thread.id, thread.id],
                    "title": "Valid thread title",
                    "category": self.other_category.id,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"category": ["Requested category could not be found."]}
        )

    @patch_category_acl({"can_merge_threads": True, "can_start_threads": False})
    def test_merge_unallowed_start_thread(self):
        """api rejects merge because category isn't allowing starting threads"""
        thread = test.post_thread(category=self.category)

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "threads": [self.thread.id, thread.id],
                    "title": "Valid thread title",
                    "category": self.category.id,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"category": ["You can't create new threads in selected category."]},
        )

    @patch_category_acl({"can_merge_threads": True})
    def test_merge_invalid_weight(self):
        """api rejects merge because final weight was invalid"""
        thread = test.post_thread(category=self.category)

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "threads": [self.thread.id, thread.id],
                    "title": "Valid thread title",
                    "category": self.category.id,
                    "weight": 4,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"weight": ["Ensure this value is less than or equal to 2."]},
        )

    @patch_category_acl({"can_merge_threads": True})
    def test_merge_unallowed_global_weight(self):
        """api rejects merge because global weight was unallowed"""
        thread = test.post_thread(category=self.category)

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "threads": [self.thread.id, thread.id],
                    "title": "Valid thread title",
                    "category": self.category.id,
                    "weight": 2,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "weight": [
                    "You don't have permission to pin threads globally "
                    "in this category."
                ]
            },
        )

    @patch_category_acl({"can_merge_threads": True})
    def test_merge_unallowed_local_weight(self):
        """api rejects merge because local weight was unallowed"""
        thread = test.post_thread(category=self.category)

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "threads": [self.thread.id, thread.id],
                    "title": "Valid thread title",
                    "category": self.category.id,
                    "weight": 1,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"weight": ["You don't have permission to pin threads in this category."]},
        )

    @patch_category_acl({"can_merge_threads": True, "can_pin_threads": 1})
    def test_merge_allowed_local_weight(self):
        """api allows local weight"""
        thread = test.post_thread(category=self.category)

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "threads": [self.thread.id, thread.id],
                    "title": "$$$",
                    "category": self.category.id,
                    "weight": 1,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "title": [
                    "Thread title should be at least 5 characters long (it has 3)."
                ]
            },
        )

    @patch_category_acl({"can_merge_threads": True, "can_pin_threads": 2})
    def test_merge_allowed_global_weight(self):
        """api allows global weight"""
        thread = test.post_thread(category=self.category)

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "threads": [self.thread.id, thread.id],
                    "title": "$$$",
                    "category": self.category.id,
                    "weight": 2,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "title": [
                    "Thread title should be at least 5 characters long (it has 3)."
                ]
            },
        )

    @patch_category_acl({"can_merge_threads": True, "can_close_threads": False})
    def test_merge_unallowed_close(self):
        """api rejects merge because closing thread was unallowed"""
        thread = test.post_thread(category=self.category)

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "threads": [self.thread.id, thread.id],
                    "title": "Valid thread title",
                    "category": self.category.id,
                    "is_closed": True,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "is_closed": [
                    "You don't have permission to close threads in this category."
                ]
            },
        )

    @patch_category_acl({"can_merge_threads": True, "can_close_threads": True})
    def test_merge_with_close(self):
        """api allows for closing thread"""
        thread = test.post_thread(category=self.category)

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "threads": [self.thread.id, thread.id],
                    "title": "$$$",
                    "category": self.category.id,
                    "weight": 0,
                    "is_closed": True,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "title": [
                    "Thread title should be at least 5 characters long (it has 3)."
                ]
            },
        )

    @patch_category_acl({"can_merge_threads": True, "can_hide_threads": 0})
    def test_merge_unallowed_hidden(self):
        """api rejects merge because hidden thread was unallowed"""
        thread = test.post_thread(category=self.category)

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "threads": [self.thread.id, thread.id],
                    "title": "Valid thread title",
                    "category": self.category.id,
                    "is_hidden": True,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "is_hidden": [
                    "You don't have permission to hide threads in this category."
                ]
            },
        )

    @patch_category_acl({"can_merge_threads": True, "can_hide_threads": 1})
    def test_merge_with_hide(self):
        """api allows for hiding thread"""
        thread = test.post_thread(category=self.category)

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "threads": [self.thread.id, thread.id],
                    "title": "$$$",
                    "category": self.category.id,
                    "weight": 0,
                    "is_hidden": True,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "title": [
                    "Thread title should be at least 5 characters long (it has 3)."
                ]
            },
        )

    @patch_category_acl({"can_merge_threads": True})
    def test_merge(self):
        """api performs basic merge"""
        posts_ids = [p.id for p in Post.objects.all()]
        thread = test.post_thread(category=self.category)

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "threads": [self.thread.id, thread.id],
                    "title": "Merged thread!",
                    "category": self.category.id,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)

        # is response json with new thread?
        response_json = response.json()

        new_thread = Thread.objects.get(pk=response_json["id"])
        new_thread.is_read = False
        new_thread.subscription = None

        user_acl = useracl.get_user_acl(self.user, cache_versions)
        add_acl_to_obj(user_acl, new_thread.category)
        add_acl_to_obj(user_acl, new_thread)

        self.assertEqual(response_json, ThreadsListSerializer(new_thread).data)

        # did posts move to new thread?
        for post in Post.objects.filter(id__in=posts_ids):
            self.assertEqual(post.thread_id, new_thread.id)

        # are old threads gone?
        self.assertEqual([t.pk for t in Thread.objects.all()], [new_thread.pk])

    @patch_category_acl(
        {
            "can_merge_threads": True,
            "can_close_threads": True,
            "can_hide_threads": 1,
            "can_pin_threads": 2,
        }
    )
    def test_merge_kitchensink(self):
        """api performs merge"""
        posts_ids = [p.id for p in Post.objects.all()]
        thread = test.post_thread(category=self.category)

        poststracker.save_read(self.user, self.thread.first_post)
        poststracker.save_read(self.user, thread.first_post)

        self.user.subscription_set.create(
            thread=self.thread,
            category=self.thread.category,
            last_read_on=self.thread.last_post_on,
            send_email=False,
        )
        self.user.subscription_set.create(
            thread=thread,
            category=thread.category,
            last_read_on=thread.last_post_on,
            send_email=False,
        )

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "threads": [self.thread.id, thread.id],
                    "title": "Merged thread!",
                    "category": self.category.id,
                    "is_closed": 1,
                    "is_hidden": 1,
                    "weight": 2,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)

        # is response json with new thread?
        response_json = response.json()

        new_thread = Thread.objects.get(pk=response_json["id"])
        new_thread.is_read = False
        new_thread.subscription = None

        self.assertEqual(new_thread.weight, 2)
        self.assertTrue(new_thread.is_closed)
        self.assertTrue(new_thread.is_hidden)

        user_acl = useracl.get_user_acl(self.user, cache_versions)
        add_acl_to_obj(user_acl, new_thread.category)
        add_acl_to_obj(user_acl, new_thread)

        self.assertEqual(response_json, ThreadsListSerializer(new_thread).data)

        # did posts move to new thread?
        for post in Post.objects.filter(id__in=posts_ids):
            self.assertEqual(post.thread_id, new_thread.id)

        # are old threads gone?
        self.assertEqual([t.pk for t in Thread.objects.all()], [new_thread.pk])

        # posts reads are kept
        postreads = self.user.postread_set.filter(post__is_event=False).order_by("id")

        self.assertEqual(
            list(postreads.values_list("post_id", flat=True)),
            [self.thread.first_post_id, thread.first_post_id],
        )
        self.assertEqual(postreads.filter(thread=new_thread).count(), 2)
        self.assertEqual(postreads.filter(category=self.category).count(), 2)

        # subscriptions are kept
        self.assertEqual(self.user.subscription_set.count(), 1)
        self.user.subscription_set.get(thread=new_thread)
        self.user.subscription_set.get(category=self.category)

    @patch_category_acl({"can_merge_threads": True})
    def test_merge_threads_merged_best_answer(self):
        """api merges two threads successfully, moving best answer to old thread"""
        other_thread = test.post_thread(self.category)

        best_answer = test.reply_thread(self.thread)
        self.thread.set_best_answer(self.user, best_answer)
        self.thread.save()

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "threads": [self.thread.id, other_thread.id],
                    "title": "Merged thread!",
                    "category": self.category.id,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)

        # best answer is set on new thread
        new_thread = Thread.objects.get(pk=response.json()["id"])
        self.assertEqual(new_thread.best_answer_id, best_answer.id)

    @patch_category_acl({"can_merge_threads": True})
    def test_merge_threads_merge_conflict_best_answer(self):
        """api errors on merge conflict, returning list of available best answers"""
        best_answer = test.reply_thread(self.thread)
        self.thread.set_best_answer(self.user, best_answer)
        self.thread.save()

        other_thread = test.post_thread(self.category)
        other_best_answer = test.reply_thread(other_thread)
        other_thread.set_best_answer(self.user, other_best_answer)
        other_thread.save()

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "threads": [self.thread.id, other_thread.id],
                    "title": "Merged thread!",
                    "category": self.category.id,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "best_answers": [
                    ["0", "Unmark all best answers"],
                    [str(self.thread.id), self.thread.title],
                    [str(other_thread.id), other_thread.title],
                ]
            },
        )

        # best answers were untouched
        self.assertEqual(self.thread.post_set.count(), 2)
        self.assertEqual(other_thread.post_set.count(), 2)
        self.assertEqual(
            Thread.objects.get(pk=self.thread.pk).best_answer_id, best_answer.id
        )
        self.assertEqual(
            Thread.objects.get(pk=other_thread.pk).best_answer_id, other_best_answer.id
        )

    @patch_category_acl({"can_merge_threads": True})
    def test_threads_merge_conflict_best_answer_invalid_resolution(self):
        """api errors on invalid merge conflict resolution"""
        best_answer = test.reply_thread(self.thread)
        self.thread.set_best_answer(self.user, best_answer)
        self.thread.save()

        other_thread = test.post_thread(self.category)
        other_best_answer = test.reply_thread(other_thread)
        other_thread.set_best_answer(self.user, other_best_answer)
        other_thread.save()

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "threads": [self.thread.id, other_thread.id],
                    "title": "Merged thread!",
                    "category": self.category.id,
                    "best_answer": other_thread.id + 10,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"best_answer": ["Invalid choice."]})

        # best answers were untouched
        self.assertEqual(self.thread.post_set.count(), 2)
        self.assertEqual(other_thread.post_set.count(), 2)
        self.assertEqual(
            Thread.objects.get(pk=self.thread.pk).best_answer_id, best_answer.id
        )
        self.assertEqual(
            Thread.objects.get(pk=other_thread.pk).best_answer_id, other_best_answer.id
        )

    @patch_category_acl({"can_merge_threads": True})
    def test_threads_merge_conflict_unmark_all_best_answers(self):
        """api unmarks all best answers when unmark all choice is selected"""
        best_answer = test.reply_thread(self.thread)
        self.thread.set_best_answer(self.user, best_answer)
        self.thread.save()

        other_thread = test.post_thread(self.category)
        other_best_answer = test.reply_thread(other_thread)
        other_thread.set_best_answer(self.user, other_best_answer)
        other_thread.save()

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "threads": [self.thread.id, other_thread.id],
                    "title": "Merged thread!",
                    "category": self.category.id,
                    "best_answer": 0,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)

        # best answer is not set on new thread
        new_thread = Thread.objects.get(pk=response.json()["id"])
        self.assertFalse(new_thread.has_best_answer)
        self.assertIsNone(new_thread.best_answer_id)

    @patch_category_acl({"can_merge_threads": True})
    def test_threads_merge_conflict_keep_first_best_answer(self):
        """api unmarks other best answer on merge"""
        best_answer = test.reply_thread(self.thread)
        self.thread.set_best_answer(self.user, best_answer)
        self.thread.save()

        other_thread = test.post_thread(self.category)
        other_best_answer = test.reply_thread(other_thread)
        other_thread.set_best_answer(self.user, other_best_answer)
        other_thread.save()

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "threads": [self.thread.id, other_thread.id],
                    "title": "Merged thread!",
                    "category": self.category.id,
                    "best_answer": self.thread.pk,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)

        # selected best answer is set on new thread
        new_thread = Thread.objects.get(pk=response.json()["id"])
        self.assertEqual(new_thread.best_answer_id, best_answer.id)

    @patch_category_acl({"can_merge_threads": True})
    def test_threads_merge_conflict_keep_other_best_answer(self):
        """api unmarks first best answer on merge"""
        best_answer = test.reply_thread(self.thread)
        self.thread.set_best_answer(self.user, best_answer)
        self.thread.save()

        other_thread = test.post_thread(self.category)
        other_best_answer = test.reply_thread(other_thread)
        other_thread.set_best_answer(self.user, other_best_answer)
        other_thread.save()

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "threads": [self.thread.id, other_thread.id],
                    "title": "Merged thread!",
                    "category": self.category.id,
                    "best_answer": other_thread.pk,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)

        # selected best answer is set on new thread
        new_thread = Thread.objects.get(pk=response.json()["id"])
        self.assertEqual(new_thread.best_answer_id, other_best_answer.id)

    @patch_category_acl({"can_merge_threads": True})
    def test_merge_threads_kept_poll(self):
        """api merges two threads successfully, keeping poll from other thread"""
        other_thread = test.post_thread(self.category)
        poll = test.post_poll(other_thread, self.user)

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "threads": [self.thread.id, other_thread.id],
                    "title": "Merged thread!",
                    "category": self.category.id,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)

        new_thread = Thread.objects.get(pk=response.json()["id"])

        # poll and its votes were kept
        self.assertEqual(Poll.objects.filter(pk=poll.pk, thread=new_thread).count(), 1)
        self.assertEqual(
            PollVote.objects.filter(poll=poll, thread=new_thread).count(), 4
        )

        self.assertEqual(Poll.objects.count(), 1)
        self.assertEqual(PollVote.objects.count(), 4)

    @patch_category_acl({"can_merge_threads": True})
    def test_merge_threads_moved_poll(self):
        """api merges two threads successfully, moving poll from old thread"""
        other_thread = test.post_thread(self.category)
        poll = test.post_poll(self.thread, self.user)

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "threads": [self.thread.id, other_thread.id],
                    "title": "Merged thread!",
                    "category": self.category.id,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)

        new_thread = Thread.objects.get(pk=response.json()["id"])

        # poll and its votes were kept
        self.assertEqual(Poll.objects.filter(pk=poll.pk, thread=new_thread).count(), 1)
        self.assertEqual(
            PollVote.objects.filter(poll=poll, thread=new_thread).count(), 4
        )

        self.assertEqual(Poll.objects.count(), 1)
        self.assertEqual(PollVote.objects.count(), 4)

    @patch_category_acl({"can_merge_threads": True})
    def test_threads_merge_conflict_poll(self):
        """api errors on merge conflict, returning list of available polls"""
        other_thread = test.post_thread(self.category)
        poll = test.post_poll(self.thread, self.user)
        other_poll = test.post_poll(other_thread, self.user)

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "threads": [self.thread.id, other_thread.id],
                    "title": "Merged thread!",
                    "category": self.category.id,
                }
            ),
            content_type="application/json",
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "polls": [
                    ["0", "Delete all polls"],
                    [
                        str(other_poll.pk),
                        "%s (%s)" % (other_poll.question, other_poll.thread.title),
                    ],
                    [str(poll.pk), "%s (%s)" % (poll.question, poll.thread.title)],
                ]
            },
        )

        # polls and votes were untouched
        self.assertEqual(Poll.objects.count(), 2)
        self.assertEqual(PollVote.objects.count(), 8)

    @patch_category_acl({"can_merge_threads": True})
    def test_threads_merge_conflict_poll_invalid_resolution(self):
        """api errors on invalid merge conflict resolution"""
        other_thread = test.post_thread(self.category)

        test.post_poll(self.thread, self.user)
        test.post_poll(other_thread, self.user)

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "threads": [self.thread.id, other_thread.id],
                    "title": "Merged thread!",
                    "category": self.category.id,
                    "poll": other_thread.poll.id + 10,
                }
            ),
            content_type="application/json",
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"poll": ["Invalid choice."]})

        # polls and votes were untouched
        self.assertEqual(Poll.objects.count(), 2)
        self.assertEqual(PollVote.objects.count(), 8)

    @patch_category_acl({"can_merge_threads": True})
    def test_threads_merge_conflict_delete_all_polls(self):
        """api deletes all polls when delete all choice is selected"""
        other_thread = test.post_thread(self.category)

        test.post_poll(self.thread, self.user)
        test.post_poll(other_thread, self.user)

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "threads": [self.thread.id, other_thread.id],
                    "title": "Merged thread!",
                    "category": self.category.id,
                    "poll": 0,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)

        # polls and votes are gone
        self.assertEqual(Poll.objects.count(), 0)
        self.assertEqual(PollVote.objects.count(), 0)

    @patch_category_acl({"can_merge_threads": True})
    def test_threads_merge_conflict_keep_first_poll(self):
        """api deletes other poll on merge"""
        other_thread = test.post_thread(self.category)
        poll = test.post_poll(self.thread, self.user)
        other_poll = test.post_poll(other_thread, self.user)

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "threads": [self.thread.id, other_thread.id],
                    "title": "Merged thread!",
                    "category": self.category.id,
                    "poll": poll.pk,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)

        # other poll and its votes are gone
        self.assertEqual(Poll.objects.count(), 1)
        self.assertEqual(PollVote.objects.count(), 4)

        Poll.objects.get(pk=poll.pk)
        with self.assertRaises(Poll.DoesNotExist):
            Poll.objects.get(pk=other_poll.pk)

    @patch_category_acl({"can_merge_threads": True})
    def test_threads_merge_conflict_keep_other_poll(self):
        """api deletes first poll on merge"""
        other_thread = test.post_thread(self.category)
        poll = test.post_poll(self.thread, self.user)
        other_poll = test.post_poll(other_thread, self.user)

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "threads": [self.thread.id, other_thread.id],
                    "title": "Merged thread!",
                    "category": self.category.id,
                    "poll": other_poll.pk,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)

        # other poll and its votes are gone
        self.assertEqual(Poll.objects.count(), 1)
        self.assertEqual(PollVote.objects.count(), 4)

        Poll.objects.get(pk=other_poll.pk)
        with self.assertRaises(Poll.DoesNotExist):
            Poll.objects.get(pk=poll.pk)



import os
from datetime import timedelta

from django.core.files import File
from django.utils import timezone

from ...categories.models import Category
from ...threads.models import AttachmentType
from ...threads.test import post_poll, post_thread
from ..audittrail import create_user_audit_trail
from ..datadownloads import (
    expire_user_data_download,
    prepare_user_data_download,
    request_user_data_download,
    user_has_data_download_request,
)
from ..models import DataDownload
from ..test import AuthenticatedUserTestCase

EXPIRATION = 4
TESTFILES_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "testfiles")
TEST_FILE_PATH = os.path.join(TESTFILES_DIR, "avatar.png")


class ExpireUserDataDownloadTests(AuthenticatedUserTestCase):
    def test_util_marks_download_as_expired(self):
        """expire_user_data_download changed data download status to expired"""
        data_download = request_user_data_download(self.user)
        data_download.status = DataDownload.STATUS_READY

        with open(TEST_FILE_PATH, "rb") as download_file:
            data_download.file = File(download_file)
            data_download.save()

        expire_user_data_download(data_download)

        self.assertEqual(data_download.status, DataDownload.STATUS_EXPIRED)

    def test_util_deletes_file(self):
        """expire_user_data_download deleted file associated with download"""
        data_download = request_user_data_download(self.user)
        data_download.status = DataDownload.STATUS_READY

        with open(TEST_FILE_PATH, "rb") as download_file:
            data_download.file = File(download_file)
            data_download.save()

        download_file_path = data_download.file.path

        expire_user_data_download(data_download)

        self.assertFalse(data_download.file)
        self.assertFalse(os.path.isdir(download_file_path))

    def test_util_expires_download_without_file(self):
        """expire_user_data_download handles missing download file"""
        data_download = request_user_data_download(self.user)
        data_download.status = DataDownload.STATUS_READY

        expire_user_data_download(data_download)

        self.assertEqual(data_download.status, DataDownload.STATUS_EXPIRED)


class PrepareUserDataDownload(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()
        self.download = request_user_data_download(self.user)

    def assert_download_is_valid(self):
        result = prepare_user_data_download(self.download, EXPIRATION)
        self.assertTrue(result)

        self.download.refresh_from_db()
        self.assertTrue(self.download.file)

    def test_prepare_basic_download(self):
        """function creates data download for basic user account"""
        self.assert_download_is_valid()

    def test_data_download_is_prepared_with_expiration_date(self):
        """function creates data download with specified expiration date"""
        expires_on = timezone.now() + timedelta(hours=EXPIRATION)
        prepare_user_data_download(self.download, EXPIRATION)
        self.download.refresh_from_db
        self.assertGreater(self.download.expires_on, expires_on)

    def test_prepare_download_with_profle_fields(self):
        """function creates data download for user with profile fields"""
        self.user.profile_fields = {"real_name": "Bob Boberthon!"}
        self.user.save()

        self.assert_download_is_valid()

    def test_prepare_download_with_tmp_avatar(self):
        """function creates data download for user with tmp avatar"""
        with open(TEST_FILE_PATH, "rb") as test_file:
            self.user.avatar_tmp = File(test_file)
            self.user.save()

        self.assert_download_is_valid()

    def test_prepare_download_with_src_avatar(self):
        """function creates data download for user with src avatar"""
        with open(TEST_FILE_PATH, "rb") as test_file:
            self.user.avatar_src = File(test_file)
            self.user.save()

        self.assert_download_is_valid()

    def test_prepare_download_with_avatar_set(self):
        """function creates data download for user with avatar set"""
        with open(TEST_FILE_PATH, "rb") as test_file:
            self.user.avatar_set.create(size=100, image=File(test_file))

        self.assert_download_is_valid()

    def test_prepare_download_with_file_attachment(self):
        """function creates data download for user with file attachment"""
        filetype = AttachmentType.objects.create(
            name="Test extension", extensions="png", mimetypes="image/png"
        )

        with open(TEST_FILE_PATH, "rb") as test_file:
            self.user.attachment_set.create(
                secret="test",
                filetype=filetype,
                uploader_name=self.user.username,
                uploader_slug=self.user.slug,
                filename="test.png",
                size=1000,
                file=File(test_file),
            )

        self.assert_download_is_valid()

    def test_prepare_download_with_image_attachment(self):
        """function creates data download for user with image attachment"""
        filetype = AttachmentType.objects.create(
            name="Test extension", extensions="png", mimetypes="image/png"
        )

        with open(TEST_FILE_PATH, "rb") as test_file:
            self.user.attachment_set.create(
                secret="test",
                filetype=filetype,
                uploader_name=self.user.username,
                uploader_slug=self.user.slug,
                filename="test.png",
                size=1000,
                image=File(test_file),
            )

        self.assert_download_is_valid()

    def test_prepare_download_with_thumbnail_attachment(self):
        """function creates data download for user with thumbnail attachment"""
        filetype = AttachmentType.objects.create(
            name="Test extension", extensions="png", mimetypes="image/png"
        )

        with open(TEST_FILE_PATH, "rb") as test_file:
            self.user.attachment_set.create(
                secret="test",
                filetype=filetype,
                uploader_name=self.user.username,
                uploader_slug=self.user.slug,
                filename="test.png",
                size=1000,
                thumbnail=File(test_file),
            )

        self.assert_download_is_valid()

    def test_prepare_download_with_self_username_change(self):
        """function creates data download for user that changed their username"""
        self.user.record_name_change(self.user, "aerith", "alice")

        self.assert_download_is_valid()

    def test_prepare_download_with_username_changed_by_staff(self):
        """function creates data download for user with username changed by staff"""
        staff_user = self.get_superuser()
        self.user.record_name_change(staff_user, "aerith", "alice")

        self.assert_download_is_valid()

    def test_prepare_download_with_username_changed_by_deleted_user(self):
        """
        function creates data download for user with username changed by deleted user
        """
        self.user.record_name_change(self.user, "aerith", "alice")
        self.user.namechanges.update(changed_by=None)

        self.assert_download_is_valid()

    def test_prepare_download_with_audit_trail(self):
        """function creates data download for user with audit trail"""
        create_user_audit_trail(self.user, "127.0.0.1", self.user)

        self.assert_download_is_valid()

    def test_prepare_download_with_post(self):
        """function creates data download for user with post"""
        category = Category.objects.get(slug="first-category")
        post_thread(category, poster=self.user)

        self.assert_download_is_valid()

    def test_prepare_download_with_owm_post_edit(self):
        """function creates data download for user with own post edit"""
        category = Category.objects.get(slug="first-category")
        thread = post_thread(category, poster=self.user)
        post = thread.first_post

        post.edits_record.create(
            category=category,
            thread=thread,
            editor=self.user,
            editor_name=self.user.username,
            editor_slug=self.user.slug,
            edited_from="edited from",
            edited_to="edited to",
        )

        self.assert_download_is_valid()

    def test_prepare_download_with_other_users_post_edit(self):
        """function creates data download for user with other user's post edit"""
        category = Category.objects.get(slug="first-category")
        thread = post_thread(category)
        post = thread.first_post

        post.edits_record.create(
            category=category,
            thread=thread,
            editor=self.user,
            editor_name=self.user.username,
            editor_slug=self.user.slug,
            edited_from="edited from",
            edited_to="edited to",
        )

        self.assert_download_is_valid()

    def test_prepare_download_with_own_post_edit_by_staff(self):
        """function creates data download for user with post edited by staff"""
        category = Category.objects.get(slug="first-category")
        thread = post_thread(category, poster=self.user)
        post = thread.first_post

        staff_user = self.get_superuser()

        post.edits_record.create(
            category=category,
            thread=thread,
            editor=staff_user,
            editor_name=staff_user.username,
            editor_slug=staff_user.slug,
            edited_from="edited from",
            edited_to="edited to",
        )

        self.assert_download_is_valid()

    def test_prepare_download_with_poll(self):
        """function creates data download for user with poll"""
        category = Category.objects.get(slug="first-category")
        thread = post_thread(category, poster=self.user)
        post_poll(thread, self.user)

        self.assert_download_is_valid()


class RequestUserDataDownloadTests(AuthenticatedUserTestCase):
    def test_util_creates_data_download_for_user_with_them_as_requester(self):
        """request_user_data_download created valid data download for user"""
        data_download = request_user_data_download(self.user)

        self.assertEqual(data_download.user, self.user)
        self.assertEqual(data_download.requester, self.user)
        self.assertEqual(data_download.requester_name, self.user.username)
        self.assertEqual(data_download.status, DataDownload.STATUS_PENDING)

    def test_util_creates_data_download_for_user_explicit_requester(self):
        """
        request_user_data_download created valid data download
        for user with other requester
        """
        requester = self.get_superuser()
        data_download = request_user_data_download(self.user, requester)

        self.assertEqual(data_download.user, self.user)
        self.assertEqual(data_download.requester, requester)
        self.assertEqual(data_download.requester_name, requester.username)
        self.assertEqual(data_download.status, DataDownload.STATUS_PENDING)


class UserHasRequestedDataDownloadTests(AuthenticatedUserTestCase):
    def test_util_returns_false_for_no_download(self):
        """
        user_has_data_download_request returns false if user has no requests in progress
        """
        self.assertFalse(user_has_data_download_request(self.user))

    def test_util_returns_false_for_ready_download(self):
        """user_has_data_download_request returns false if user has ready download"""
        data_download = request_user_data_download(self.user)
        data_download.status = DataDownload.STATUS_READY
        data_download.save()

        self.assertFalse(user_has_data_download_request(self.user))

    def test_util_returns_false_for_expired_download(self):
        """user_has_data_download_request returns false if user has expired download"""
        data_download = request_user_data_download(self.user)
        data_download.status = DataDownload.STATUS_EXPIRED
        data_download.save()

        self.assertFalse(user_has_data_download_request(self.user))

    def test_util_returns_true_for_pending_download(self):
        """user_has_data_download_request returns true if user has pending download"""
        data_download = request_user_data_download(self.user)
        data_download.status = DataDownload.STATUS_PENDING
        data_download.save()

        self.assertTrue(user_has_data_download_request(self.user))

    def test_util_returns_true_for_processing_download(self):
        """
        user_has_data_download_request returns true if user has processing download
        """
        data_download = request_user_data_download(self.user)
        data_download.status = DataDownload.STATUS_PROCESSING
        data_download.save()

        self.assertTrue(user_has_data_download_request(self.user))



from django import forms
from django.contrib.auth import get_user_model
from django.contrib.auth.password_validation import validate_password
from django.core.exceptions import ValidationError
from django.utils.translation import gettext as _

from ..bans import get_email_ban, get_ip_ban, get_username_ban
from ..validators import validate_email, validate_new_registration, validate_username

User = get_user_model()


class BaseRegisterForm(forms.Form):
    username = forms.CharField()
    email = forms.CharField(validators=[validate_email])

    terms_of_service = forms.IntegerField(required=False)
    privacy_policy = forms.IntegerField(required=False)

    def __init__(self, *args, **kwargs):
        self.agreements = kwargs.pop("agreements")
        self.request = kwargs.pop("request")
        super().__init__(*args, **kwargs)

    def clean_username(self):
        data = self.cleaned_data["username"]

        validate_username(self.request.settings, data)
        ban = get_username_ban(data, registration_only=True)
        if ban:
            if ban.user_message:
                raise ValidationError(ban.user_message)
            else:
                raise ValidationError(_("This username is not allowed."))
        return data

    def clean_email(self):
        data = self.cleaned_data["email"]

        ban = get_email_ban(data, registration_only=True)
        if ban:
            if ban.user_message:
                raise ValidationError(ban.user_message)
            else:
                raise ValidationError(_("This e-mail address is not allowed."))
        return data

    def clean_agreements(self, data):
        for field_name, agreement in self.agreements.items():
            if data.get(field_name) != agreement["id"]:
                error = ValueError(_("This agreement is required."))
                self.add_error(field_name, error)

    def raise_if_ip_banned(self):
        ban = get_ip_ban(self.request.user_ip, registration_only=True)
        if ban:
            if ban.user_message:
                raise ValidationError(ban.user_message)
            else:
                raise ValidationError(
                    _("New registrations from this IP address are not allowed.")
                )


class SocialAuthRegisterForm(BaseRegisterForm):
    def clean(self):
        cleaned_data = super().clean()

        self.clean_agreements(cleaned_data)
        self.raise_if_ip_banned()

        validate_new_registration(self.request, cleaned_data, self.add_error)

        return cleaned_data


class RegisterForm(BaseRegisterForm):
    password = forms.CharField(strip=False)

    # placeholder field for setting captcha errors on form
    captcha = forms.CharField(required=False)

    def full_clean_password(self, cleaned_data):
        if cleaned_data.get("password"):
            validate_password(
                cleaned_data["password"],
                user=User(
                    username=cleaned_data.get("username"),
                    email=cleaned_data.get("email"),
                ),
            )

    def clean(self):
        cleaned_data = super().clean()

        self.clean_agreements(cleaned_data)
        self.raise_if_ip_banned()

        try:
            self.full_clean_password(cleaned_data)
        except forms.ValidationError as e:
            self.add_error("password", e)

        validate_new_registration(self.request, cleaned_data, self.add_error)

        return cleaned_data



from django import forms
from django.utils.translation import gettext_lazy as _

from ....admin.forms import ColorField, YesNoSwitch
from ...models import SocialAuthProvider
from ..ordering import get_next_free_order


class ProviderForm(forms.ModelForm):
    button_text = forms.CharField(label=_("Button text"), required=False)
    button_color = ColorField(label=_("Button color"), required=False)
    is_active = YesNoSwitch(label=_("Enable this provider"))

    class Meta:
        model = SocialAuthProvider
        fields = ["button_text", "button_color", "is_active"]

    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop("request")
        super().__init__(*args, **kwargs)

    def save(self):
        settings = {}
        for setting, value in self.cleaned_data.items():
            if setting not in ["button_text", "button_color", "is_active"]:
                settings[setting] = value
        self.instance.settings = settings

        if "is_active" in self.changed_data and self.cleaned_data.get("is_active"):
            self.instance.order = get_next_free_order()

        self.instance.save()


class OAuthProviderForm(ProviderForm):
    associate_by_email = YesNoSwitch(
        label=_("Associate existing users by email"),
        help_text=_(
            "Enabling this option will result in the user being signed in if their "
            "forum account was registered using same e-mail address as used on the "
            "social site. Otherwise they will be asked to specify a different e-mail "
            "to continue sign in using social site."
        ),
    )

    def clean(self):
        cleaned_data = super().clean()

        if cleaned_data.get("is_active"):
            if not cleaned_data.get("key"):
                self.add_error("key", _("This field is required."))
            if not cleaned_data.get("secret"):
                self.add_error("secret", _("This field is required."))

        return cleaned_data



from django.core.exceptions import PermissionDenied
from django.http import Http404
from django.test import TestCase

from ..apipatch import ApiPatch, InvalidAction


class MockRequest:
    def __init__(self, data=None):
        self.data = data


class MockObject:
    def __init__(self, pk):
        self.id = pk
        self.pk = pk


class ApiPatchTests(TestCase):
    def test_add(self):
        """add method adds function to patch object"""
        patch = ApiPatch()

        def mock_function():
            pass

        patch.add("test-add", mock_function)

        self.assertEqual(len(patch._actions), 1)
        self.assertEqual(patch._actions[0]["op"], "add")
        self.assertEqual(patch._actions[0]["path"], "test-add")
        self.assertEqual(patch._actions[0]["handler"], mock_function)

    def test_remove(self):
        """remove method adds function to patch object"""
        patch = ApiPatch()

        def mock_function():
            pass

        patch.remove("test-remove", mock_function)

        self.assertEqual(len(patch._actions), 1)
        self.assertEqual(patch._actions[0]["op"], "remove")
        self.assertEqual(patch._actions[0]["path"], "test-remove")
        self.assertEqual(patch._actions[0]["handler"], mock_function)

    def test_replace(self):
        """replace method adds function to patch object"""
        patch = ApiPatch()

        def mock_function():
            pass

        patch.replace("test-replace", mock_function)

        self.assertEqual(len(patch._actions), 1)
        self.assertEqual(patch._actions[0]["op"], "replace")
        self.assertEqual(patch._actions[0]["path"], "test-replace")
        self.assertEqual(patch._actions[0]["handler"], mock_function)

    def test_validate_action(self):
        """validate_action method validates action dict"""
        patch = ApiPatch()

        VALID_ACTIONS = [
            {"op": "add", "path": "test", "value": 42},
            {"op": "remove", "path": "other-test", "value": "Lorem"},
            {"op": "replace", "path": "false-test", "value": None},
        ]

        for action in VALID_ACTIONS:
            patch.validate_action(action)

        # undefined op
        UNSUPPORTED_ACTIONS = ({}, {"op": ""}, {"no": "op"})

        for action in UNSUPPORTED_ACTIONS:
            try:
                patch.validate_action(action)
            except InvalidAction as e:
                self.assertEqual(e.args[0], "undefined op")

        # unsupported op
        try:
            patch.validate_action({"op": "nope"})
        except InvalidAction as e:
            self.assertEqual(e.args[0], '"nope" op is unsupported')

        # op lacking patch
        try:
            patch.validate_action({"op": "add"})
        except InvalidAction as e:
            self.assertEqual(e.args[0], '"add" op has to specify path')

        # op lacking value
        try:
            patch.validate_action({"op": "add", "path": "yolo"})
        except InvalidAction as e:
            self.assertEqual(e.args[0], '"add" op has to specify value')

        # empty value is allowed
        try:
            patch.validate_action({"op": "add", "path": "yolo", "value": ""})
        except InvalidAction as e:
            self.assertEqual(e.args[0], '"add" op has to specify value')

    def test_dispatch_action(self):
        """dispatch_action calls specified actions"""
        patch = ApiPatch()

        mock_target = MockObject(13)

        def action_a(request, target, value):
            self.assertEqual(request, "request")
            self.assertEqual(target, mock_target)
            return {"a": value * 2, "b": 111}

        patch.replace("abc", action_a)

        def action_b(request, target, value):
            self.assertEqual(request, "request")
            self.assertEqual(target, mock_target)
            return {"b": value * 10}

        patch.replace("abc", action_b)

        def action_fail(request, target, value):
            self.fail("unrequired action was called")

        patch.add("c", action_fail)
        patch.remove("c", action_fail)
        patch.replace("c", action_fail)

        patch_dict = {"id": 123}

        patch.dispatch_action(
            patch_dict,
            "request",
            mock_target,
            {"op": "replace", "path": "abc", "value": 5},
        )

        self.assertEqual(len(patch_dict), 3)
        self.assertEqual(patch_dict["id"], 123)
        self.assertEqual(patch_dict["a"], 10)
        self.assertEqual(patch_dict["b"], 50)

    def test_dispatch(self):
        """dispatch calls actions and returns response"""
        patch = ApiPatch()

        def action_error(request, target, value):
            if value == "404":
                raise Http404()
            if value == "perm":
                raise PermissionDenied("yo ain't doing that!")

        patch.replace("error", action_error)

        def action_mutate(request, target, value):
            return {"value": value * 2}

        patch.replace("mutate", action_mutate)

        # dispatch requires list as an argument
        response = patch.dispatch(MockRequest({}), {})
        self.assertEqual(response.status_code, 400)

        self.assertEqual(
            response.data["detail"], "PATCH request should be list of operations"
        )

        # valid dispatch
        response = patch.dispatch(
            MockRequest(
                [
                    {"op": "replace", "path": "mutate", "value": 2},
                    {"op": "replace", "path": "mutate", "value": 6},
                    {"op": "replace", "path": "mutate", "value": 7},
                ]
            ),
            MockObject(13),
        )

        self.assertEqual(response.status_code, 200)

        self.assertEqual(len(response.data["detail"]), 3)
        self.assertEqual(response.data["detail"][0], "ok")
        self.assertEqual(response.data["detail"][1], "ok")
        self.assertEqual(response.data["detail"][2], "ok")
        self.assertEqual(response.data["id"], 13)
        self.assertEqual(response.data["value"], 14)

        # invalid action in dispatch
        response = patch.dispatch(
            MockRequest(
                [
                    {"op": "replace", "path": "mutate", "value": 2},
                    {"op": "replace", "path": "mutate", "value": 6},
                    {"op": "replace"},
                    {"op": "replace", "path": "mutate", "value": 7},
                ]
            ),
            MockObject(13),
        )

        self.assertEqual(response.status_code, 400)

        self.assertEqual(len(response.data["detail"]), 3)
        self.assertEqual(response.data["detail"][0], "ok")
        self.assertEqual(response.data["detail"][1], "ok")
        self.assertEqual(response.data["detail"][2], '"replace" op has to specify path')
        self.assertEqual(response.data["id"], 13)
        self.assertEqual(response.data["value"], 12)

        # action in dispatch raised 404
        response = patch.dispatch(
            MockRequest(
                [
                    {"op": "replace", "path": "mutate", "value": 2},
                    {"op": "replace", "path": "error", "value": "404"},
                    {"op": "replace", "path": "mutate", "value": 6},
                    {"op": "replace", "path": "mutate", "value": 7},
                ]
            ),
            MockObject(13),
        )

        self.assertEqual(response.status_code, 400)

        self.assertEqual(len(response.data["detail"]), 2)
        self.assertEqual(response.data["detail"][0], "ok")
        self.assertEqual(response.data["detail"][1], "NOT FOUND")
        self.assertEqual(response.data["id"], 13)
        self.assertEqual(response.data["value"], 4)

        # action in dispatch raised perm denied
        response = patch.dispatch(
            MockRequest(
                [
                    {"op": "replace", "path": "mutate", "value": 2},
                    {"op": "replace", "path": "mutate", "value": 6},
                    {"op": "replace", "path": "mutate", "value": 9},
                    {"op": "replace", "path": "error", "value": "perm"},
                ]
            ),
            MockObject(13),
        )

        self.assertEqual(response.status_code, 400)

        self.assertEqual(len(response.data["detail"]), 4)
        self.assertEqual(response.data["detail"][0], "ok")
        self.assertEqual(response.data["detail"][1], "ok")
        self.assertEqual(response.data["detail"][2], "ok")
        self.assertEqual(response.data["detail"][3], "yo ain't doing that!")
        self.assertEqual(response.data["id"], 13)
        self.assertEqual(response.data["value"], 18)



# Generated by Django 1.11.5 on 2017-10-07 14:49
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("misago_readtracker", "0002_postread")]

    operations = []



from django.urls import include, path

# Setup Django admin to work with Misago auth
from django.contrib import admin
from django.utils import timezone
from django.views.decorators.cache import cache_page
from django.views.decorators.http import last_modified
from django.views.i18n import JavaScriptCatalog

from . import views
from ...users.forms.auth import AdminAuthenticationForm

admin.autodiscover()
admin.site.login_form = AdminAuthenticationForm

urlpatterns = [
    path("", include("social_django.urls", namespace="social")),
    path("forum/", include("misago.urls", namespace="misago")),
    path("django-admin/", admin.site.urls),
    path(
        "django-i18n.js",
        cache_page(86400 * 2, key_prefix="misagojsi18n")(
            last_modified(lambda req, **kw: timezone.now())(
                JavaScriptCatalog.as_view(packages=["misago"])
            )
        ),
        name="django-i18n",
    ),
    path("forum/test-pagination/", views.test_pagination, name="test-pagination"),
    path(
        "forum/test-pagination/<int:page>/",
        views.test_pagination,
        name="test-pagination",
    ),
    path(
        "forum/test-paginated-response/",
        views.test_paginated_response,
        name="test-paginated-response",
    ),
    path(
        "forum/test-paginated-response-data/",
        views.test_paginated_response_data,
        name="test-paginated-response-data",
    ),
    path(
        "forum/test-paginated-response-serializer/",
        views.test_paginated_response_serializer,
        name="test-paginated-response-serializer",
    ),
    path(
        "forum/test-paginated-response-data-serializer/",
        views.test_paginated_response_data_serializer,
        name="test-paginated-response-data-serializer",
    ),
    path(
        "forum/test-paginated-response-data-extra/",
        views.test_paginated_response_data_extra,
        name="test-paginated-response-data-extra",
    ),
    path(
        "forum/test-valid-slug/<slug:slug>-<int:pk>/",
        views.validate_slug_view,
        name="validate-slug-view",
    ),
    path("forum/test-banned/", views.raise_misago_banned, name="raise-misago-banned"),
    path("forum/test-403/", views.raise_misago_403, name="raise-misago-403"),
    path("forum/test-404/", views.raise_misago_404, name="raise-misago-404"),
    path("forum/test-405/", views.raise_misago_405, name="raise-misago-405"),
    path(
        "forum/social-auth-failed/",
        views.raise_social_auth_failed,
        name="raise-social-auth-failed",
    ),
    path(
        "forum/social-wrong-backend/",
        views.raise_social_wrong_backend,
        name="raise-social-wrong-backend",
    ),
    path(
        "forum/social-not-allowed-to-disconnect/",
        views.raise_social_not_allowed_to_disconnect,
        name="raise-social-not-allowed-to-disconnect",
    ),
    path(
        "forum/raise-social-auth-failed-message/",
        views.raise_social_auth_failed_message,
        name="raise-social-auth-failed-message",
    ),
    path(
        "forum/raise-social-auth-banned/",
        views.raise_social_auth_banned,
        name="raise-social-auth-banned",
    ),
    path("test-403/", views.raise_403, name="raise-403"),
    path("test-404/", views.raise_404, name="raise-404"),
    path("test-redirect/", views.test_redirect, name="test-redirect"),
    path("test-require-post/", views.test_require_post, name="test-require-post"),
]



from django.http import Http404, HttpResponse


def google_site_verification(request, token):
    if token != request.settings.google_site_verification:
        raise Http404()

    return HttpResponse("google-site-verification: google%s.html" % token)



from django.utils import timezone

from ..readtracker import poststracker
from .models import Post


def record_event(request, thread, event_type, context=None, commit=True):
    time_now = timezone.now()

    event = Post.objects.create(
        category=thread.category,
        thread=thread,
        poster=request.user,
        poster_name=request.user.username,
        original="-",
        parsed="-",
        posted_on=time_now,
        updated_on=time_now,
        is_event=True,
        event_type=event_type,
        event_context=context,
    )

    thread.has_events = True
    thread.set_last_post(event)
    if commit:
        thread.save()

    if not (thread.is_hidden and thread.is_unapproved):
        thread.category.set_last_thread(thread)
        if commit:
            thread.category.save()

    poststracker.save_read(request.user, event)

    return event



# Generated by Django 1.11.13 on 2018-06-03 18:46
import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("misago_users", "0011_auto_20180331_2208"),
    ]

    operations = [
        migrations.CreateModel(
            name="AuditTrail",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("object_id", models.PositiveIntegerField()),
                (
                    "created_on",
                    models.DateTimeField(
                        db_index=True, default=django.utils.timezone.now
                    ),
                ),
                ("ip_address", models.GenericIPAddressField()),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.ContentType",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={"ordering": ["-pk"]},
        )
    ]



from ..parser import parse


def test_single_line_spoiler(request_mock, user, snapshot):
    text = "[spoiler]Daenerys and Jon live happily ever after![/spoiler]"
    result = parse(text, request_mock, user)
    snapshot.assert_match(result["parsed_text"])


def test_spoiler_can_contain_bbcode_or_markdown(request_mock, user, snapshot):
    text = "[spoiler]Sit **amet** [u]elit[/u].[/spoiler]"
    result = parse(text, request_mock, user)
    snapshot.assert_match(result["parsed_text"])


def test_multi_line_spoiler(request_mock, user, snapshot):
    text = """
[spoiler]
Sit amet elit.

Another line.
[/spoiler]
"""
    result = parse(text, request_mock, user)
    snapshot.assert_match(result["parsed_text"])


def test_spoilers_can_be_nested(request_mock, user, snapshot):
    text = """
[spoiler]
Sit amet elit.
[spoiler]Nested spoiler[/spoiler]
[/spoiler]
"""
    result = parse(text, request_mock, user)
    snapshot.assert_match(result["parsed_text"])


# Regression test for weird edge case in which hr gets moved outside of spoiler
def test_spoilers_can_contain_hr_markdown(request_mock, user, snapshot):
    text = """
[spoiler]
Sit amet elit.
- - - - -
Another line.
[/spoiler]
"""
    result = parse(text, request_mock, user)
    snapshot.assert_match(result["parsed_text"])



import re

from django.forms import ValidationError
from django.utils.translation import gettext as _

FILENAME_CONTENT = re.compile(r"([a-zA-Z0-9]|\.|_|-)+")
FILENAME_TANGIBILITY = re.compile(r"[a-zA-Z0-9]")


def validate_css_name(filename):
    if not filename.lower().endswith(".css"):
        raise ValidationError(_("Name is missing an .css extension."))

    if filename.startswith("."):
        raise ValidationError(_('Name can\'t start with period (".").'))

    if not FILENAME_CONTENT.fullmatch(filename[:-4]):
        raise ValidationError(
            _(
                "Name can contain only latin alphabet characters, "
                "digits, dots, underscores and dashes."
            )
        )

    if not FILENAME_TANGIBILITY.match(filename[:-4]):
        raise ValidationError(
            _("Name has to contain at least one latin alphabet character or digit.")
        )


def validate_css_name_is_available(instance, name):
    queryset = instance.theme.css.filter(name=name)
    if instance.pk:
        queryset = queryset.exclude(pk=instance.pk)
    if queryset.exists():
        raise ValidationError(_("This name is already in use by other asset."))



import logging

from django.core.management.base import BaseCommand
from django.utils.translation import gettext

from ....conf import settings
from ....conf.shortcuts import get_dynamic_settings
from ....core.mail import mail_user
from ....core.pgutils import chunk_queryset
from ...datadownloads import prepare_user_data_download
from ...models import DataDownload

logger = logging.getLogger("misago.users.datadownloads")


class Command(BaseCommand):
    help = "Prepares user data downloads."
    leave_locale_alone = True

    def handle(self, *args, **options):
        working_dir = settings.MISAGO_USER_DATA_DOWNLOADS_WORKING_DIR
        if not working_dir:
            self.stdout.write(
                "MISAGO_USER_DATA_DOWNLOADS_WORKING_DIR has to be set in order for "
                "this feature to work."
            )
            return

        dynamic_settings = get_dynamic_settings()
        expires_in = dynamic_settings.data_downloads_expiration

        downloads_prepared = 0
        queryset = DataDownload.objects.select_related("user")
        queryset = queryset.filter(status=DataDownload.STATUS_PENDING)
        for data_download in chunk_queryset(queryset):
            if prepare_user_data_download(data_download, expires_in, logger):
                user = data_download.user
                subject = gettext("%(user)s, your data download is ready") % {
                    "user": user
                }
                mail_user(
                    user,
                    subject,
                    "misago/emails/data_download",
                    context={
                        "data_download": data_download,
                        "expires_in": expires_in,
                        "settings": dynamic_settings,
                    },
                )

                downloads_prepared += 1

        self.stdout.write("Data downloads prepared: %s" % downloads_prepared)



from .audittrail import create_user_audit_trail
from .avatars import set_default_avatar, set_default_avatar_from_url
from .models import User


def setup_new_user(settings, user, *, avatar_url=None):
    set_default_subscription_options(settings, user)

    if avatar_url:
        set_default_avatar_from_url(user, avatar_url)
    else:
        set_default_avatar(
            user, settings.default_avatar, settings.default_gravatar_fallback
        )

    if user.joined_from_ip:
        create_user_audit_trail(user, user.joined_from_ip, user)


SUBSCRIPTION_CHOICES = {
    "no": User.SUBSCRIPTION_NONE,
    "watch": User.SUBSCRIPTION_NOTIFY,
    "watch_email": User.SUBSCRIPTION_ALL,
}


def set_default_subscription_options(settings, user):
    started_threads = SUBSCRIPTION_CHOICES[settings.subscribe_start]
    user.subscribe_to_started_threads = started_threads

    replied_threads = SUBSCRIPTION_CHOICES[settings.subscribe_reply]
    user.subscribe_to_replied_threads = replied_threads

    user.save(
        update_fields=["subscribe_to_replied_threads", "subscribe_to_replied_threads"]
    )



from io import StringIO

from django.core.management import call_command
from django.test import TestCase

from ...acl import ACL_CACHE
from ...cache.test import assert_invalidates_cache
from ..management.commands import fixcategoriestree
from ..models import Category


def run_command():
    """Run the management command"""
    command = fixcategoriestree.Command()
    out = StringIO()
    call_command(command, stdout=out)


class FixCategoriesTreeTests(TestCase):
    """
    The purpose is the verify that the management command
    fixes the lft/rght values of the thread category tree.
    """

    def setUp(self):
        Category.objects.create(
            name="Test", slug="test", parent=Category.objects.root_category()
        )
        self.fetch_categories()

    def assertValidTree(self, expected_tree):
        root = Category.objects.root_category()
        queryset = Category.objects.filter(tree_id=root.tree_id).order_by("lft")

        current_tree = []
        for category in queryset:
            current_tree.append(
                (category, category.get_level(), category.lft, category.rght)
            )

        for i, category in enumerate(expected_tree):
            _category = current_tree[i]
            if category[0] != _category[0]:
                self.fail(
                    ("expected category at index #%s to be %s, " "found %s instead")
                    % (i, category[0], _category[0])
                )
            if category[1] != _category[1]:
                self.fail(
                    ("expected level at index #%s to be %s, " "found %s instead")
                    % (i, category[1], _category[1])
                )
            if category[2] != _category[2]:
                self.fail(
                    ("expected lft at index #%s to be %s, " "found %s instead")
                    % (i, category[2], _category[2])
                )
            if category[3] != _category[3]:
                self.fail(
                    ("expected lft at index #%s to be %s, " "found %s instead")
                    % (i, category[3], _category[3])
                )

    def fetch_categories(self):
        """gets a fresh version from the database"""
        self.root = Category.objects.root_category()
        self.first_category = Category.objects.get(slug="first-category")
        self.test_category = Category.objects.get(slug="test")

    def test_fix_categories_tree_unaffected(self):
        """Command should not affect a healthy three"""
        tree_id = self.root.tree_id
        run_command()

        self.fetch_categories()

        self.assertValidTree(
            [
                (self.root, 0, 1, 6),
                (self.first_category, 1, 2, 3),
                (self.test_category, 1, 4, 5),
            ]
        )

        self.assertEqual(self.root.tree_id, tree_id, msg="tree_id changed by command")

    def test_fix_categories_tree_affected(self):
        """Command should fix a broken tree"""
        # Root node with too narrow lft/rght range
        Category.objects.filter(id=self.root.id).update(lft=1, rght=4)
        # Make conflicting/identical lft/rght range
        Category.objects.filter(id=self.test_category.id).update(lft=2, rght=3)

        run_command()
        self.fetch_categories()

        self.assertValidTree(
            [
                (self.root, 0, 1, 6),
                (self.test_category, 1, 2, 3),
                (self.first_category, 1, 4, 5),
            ]
        )

    def test_fixing_categories_tree_invalidates_acl_cache(self):
        with assert_invalidates_cache(ACL_CACHE):
            run_command()



from importlib import import_module

from django.apps import apps

from .site import site
from .urlpatterns import urlpatterns


def discover_misago_admin():
    for app in apps.get_app_configs():
        module = import_module(app.name)
        if not hasattr(module, "admin"):
            continue

        admin_module = import_module("%s.admin" % app.name)
        if hasattr(admin_module, "MisagoAdminExtension"):
            extension = getattr(admin_module, "MisagoAdminExtension")()
            if hasattr(extension, "register_navigation_nodes"):
                extension.register_navigation_nodes(site)
            if hasattr(extension, "register_urlpatterns"):
                extension.register_urlpatterns(urlpatterns)



from datetime import timedelta

import pytest
from ariadne import gql
from django.utils import timezone

from ....threads.models import Attachment, AttachmentType
from ....threads.test import post_thread
from ....users.datadownloads import request_user_data_download
from ....users.deletesrecord import record_user_deleted_by_self
from ....users.test import create_test_user
from ..analytics import cumulate_data


test_query = gql(
    """
        query getAnalytics($span: Int!) {
            analytics(span: $span) {
                users {
                    ...data
                }
                userDeletions {
                    ...data
                }
                threads {
                    ...data
                }
                posts {
                    ...data
                }
                attachments {
                    ...data
                }
                dataDownloads {
                    ...data
                }
            }
        }

        fragment data on AnalyticsData {
            current
            currentCumulative
            previous
            previousCumulative
        }
    """
)


previous_datetime = timezone.now() - timedelta(days=30)
excluded_datetime = timezone.now() - timedelta(days=60)


def test_query_without_data_executes_without_errors(admin_graphql_client):
    result = admin_graphql_client.query(test_query, {"span": 30})
    assert result["analytics"]


def test_all_analytics_are_limited_to_requested_span(admin_graphql_client):
    result = admin_graphql_client.query(test_query, {"span": 30})
    for model_analytics in result["analytics"].values():
        assert len(model_analytics["current"]) == 30
        assert len(model_analytics["currentCumulative"]) == 30
        assert len(model_analytics["previous"]) == 30
        assert len(model_analytics["previousCumulative"]) == 30


def test_large_analytics_span_is_reduced_to_360(admin_graphql_client):
    result = admin_graphql_client.query(test_query, {"span": 3000})
    for model_analytics in result["analytics"].values():
        assert len(model_analytics["current"]) == 360
        assert len(model_analytics["currentCumulative"]) == 360
        assert len(model_analytics["previous"]) == 360
        assert len(model_analytics["previousCumulative"]) == 360


def test_short_analytics_span_is_extended_to_30(admin_graphql_client):
    result = admin_graphql_client.query(test_query, {"span": 0})
    for model_analytics in result["analytics"].values():
        assert len(model_analytics["current"]) == 30
        assert len(model_analytics["currentCumulative"]) == 30
        assert len(model_analytics["previous"]) == 30
        assert len(model_analytics["previousCumulative"]) == 30


def test_recent_user_registration_appears_in_current_analytics(admin_graphql_client):
    create_test_user("User", "user@example.com")
    result = admin_graphql_client.query(test_query, {"span": 30})
    analytics = result["analytics"]["users"]
    assert sum(analytics["current"]) == 2  # includes admin
    assert sum(analytics["previous"]) == 0


def test_older_user_registration_appears_in_previous_analytics(admin_graphql_client):
    create_test_user("User", "user@example.com", joined_on=previous_datetime)
    result = admin_graphql_client.query(test_query, {"span": 30})
    analytics = result["analytics"]["users"]
    assert sum(analytics["current"]) == 1  # includes admin
    assert sum(analytics["previous"]) == 1


def test_old_user_registration_is_excluded_from_analytics(admin_graphql_client):
    create_test_user("User", "user@example.com", joined_on=excluded_datetime)
    result = admin_graphql_client.query(test_query, {"span": 30})
    analytics = result["analytics"]["users"]
    assert sum(analytics["current"]) == 1  # includes admin
    assert sum(analytics["previous"]) == 0


def test_recent_thread_appears_in_current_analytics(
    admin_graphql_client, default_category
):
    post_thread(default_category)
    result = admin_graphql_client.query(test_query, {"span": 30})
    analytics = result["analytics"]["threads"]
    assert sum(analytics["current"]) == 1
    assert sum(analytics["previous"]) == 0


def test_older_thread_appears_in_previous_analytics(
    admin_graphql_client, default_category
):
    post_thread(default_category, started_on=previous_datetime)
    result = admin_graphql_client.query(test_query, {"span": 30})
    analytics = result["analytics"]["threads"]
    assert sum(analytics["current"]) == 0
    assert sum(analytics["previous"]) == 1


def test_old_thread_is_excluded_from_analytics(admin_graphql_client, default_category):
    post_thread(default_category, started_on=excluded_datetime)
    result = admin_graphql_client.query(test_query, {"span": 30})
    analytics = result["analytics"]["threads"]
    assert sum(analytics["current"]) == 0
    assert sum(analytics["previous"]) == 0


def test_recent_post_appears_in_current_analytics(
    admin_graphql_client, default_category
):
    post_thread(default_category)
    result = admin_graphql_client.query(test_query, {"span": 30})
    analytics = result["analytics"]["posts"]
    assert sum(analytics["current"]) == 1
    assert sum(analytics["previous"]) == 0


def test_older_post_appears_in_previous_analytics(
    admin_graphql_client, default_category
):
    post_thread(default_category, started_on=previous_datetime)
    result = admin_graphql_client.query(test_query, {"span": 30})
    analytics = result["analytics"]["posts"]
    assert sum(analytics["current"]) == 0
    assert sum(analytics["previous"]) == 1


def test_old_post_is_excluded_from_analytics(admin_graphql_client, default_category):
    post_thread(default_category, started_on=excluded_datetime)
    result = admin_graphql_client.query(test_query, {"span": 30})
    analytics = result["analytics"]["posts"]
    assert sum(analytics["current"]) == 0
    assert sum(analytics["previous"]) == 0


@pytest.fixture
def attachment_type(db):
    return AttachmentType.objects.create(name="test", extensions="test")


def test_recent_attachment_appears_in_current_analytics(
    admin_graphql_client, attachment_type
):
    Attachment.objects.create(
        filetype=attachment_type,
        uploader_name="test",
        uploader_slug="test",
        filename="test",
    )

    result = admin_graphql_client.query(test_query, {"span": 30})
    analytics = result["analytics"]["attachments"]
    assert sum(analytics["current"]) == 1
    assert sum(analytics["previous"]) == 0


def test_older_attachment_appears_in_previous_analytics(
    admin_graphql_client, attachment_type
):
    Attachment.objects.create(
        filetype=attachment_type,
        uploader_name="test",
        uploader_slug="test",
        filename="test",
        uploaded_on=previous_datetime,
    )

    result = admin_graphql_client.query(test_query, {"span": 30})
    analytics = result["analytics"]["attachments"]
    assert sum(analytics["current"]) == 0
    assert sum(analytics["previous"]) == 1


def test_old_attachment_is_excluded_from_analytics(
    admin_graphql_client, attachment_type
):
    Attachment.objects.create(
        filetype=attachment_type,
        uploader_name="test",
        uploader_slug="test",
        filename="test",
        uploaded_on=excluded_datetime,
    )

    result = admin_graphql_client.query(test_query, {"span": 30})
    analytics = result["analytics"]["attachments"]
    assert sum(analytics["current"]) == 0
    assert sum(analytics["previous"]) == 0


def test_recent_data_download_appears_in_current_analytics(
    admin_graphql_client, superuser
):
    request_user_data_download(superuser)

    result = admin_graphql_client.query(test_query, {"span": 30})
    analytics = result["analytics"]["dataDownloads"]
    assert sum(analytics["current"]) == 1
    assert sum(analytics["previous"]) == 0


def test_older_data_download_appears_in_previous_analytics(
    admin_graphql_client, superuser
):
    download = request_user_data_download(superuser)
    download.requested_on = previous_datetime
    download.save()

    result = admin_graphql_client.query(test_query, {"span": 30})
    analytics = result["analytics"]["dataDownloads"]
    assert sum(analytics["current"]) == 0
    assert sum(analytics["previous"]) == 1


def test_old_data_download_is_excluded_from_analytics(admin_graphql_client, superuser):
    download = request_user_data_download(superuser)
    download.requested_on = excluded_datetime
    download.save()

    result = admin_graphql_client.query(test_query, {"span": 30})
    analytics = result["analytics"]["dataDownloads"]
    assert sum(analytics["current"]) == 0
    assert sum(analytics["previous"]) == 0


def test_recent_user_deletion_appears_in_current_analytics(admin_graphql_client):
    record_user_deleted_by_self()
    result = admin_graphql_client.query(test_query, {"span": 30})
    analytics = result["analytics"]["userDeletions"]
    assert sum(analytics["current"]) == 1
    assert sum(analytics["previous"]) == 0


def test_older_user_deletion_appears_in_previous_analytics(admin_graphql_client):
    deletion = record_user_deleted_by_self()
    deletion.deleted_on = previous_datetime
    deletion.save()

    result = admin_graphql_client.query(test_query, {"span": 30})
    analytics = result["analytics"]["userDeletions"]
    assert sum(analytics["current"]) == 0
    assert sum(analytics["previous"]) == 1


def test_old_user_deletion_is_excluded_from_analytics(admin_graphql_client):
    deletion = record_user_deleted_by_self()
    deletion.deleted_on = excluded_datetime
    deletion.save()

    result = admin_graphql_client.query(test_query, {"span": 30})
    analytics = result["analytics"]["userDeletions"]
    assert sum(analytics["current"]) == 0
    assert sum(analytics["previous"]) == 0


def test_data_is_cumulated():
    assert cumulate_data([1, 2]) == [1, 3]



# Generated by Django 1.11.9 on 2018-03-18 20:40
from django.db import migrations

_ = lambda s: s


def create_default_categories_roles(apps, schema_editor):
    CategoryRole = apps.get_model("misago_categories", "CategoryRole")

    CategoryRole.objects.create(
        name=_("Q&A user"),
        permissions={
            # best answers perms
            "misago.threads.permissions.bestanswers": {
                "can_mark_best_answers": 1,
                "can_change_marked_answers": 1,
                "best_answer_change_time": 60 * 36,  # 1.5 day
            }
        },
    )

    CategoryRole.objects.create(
        name=_("Q&A moderator"),
        permissions={
            # best answers perms
            "misago.threads.permissions.bestanswers": {
                "can_mark_best_answers": 2,
                "can_change_marked_answers": 2,
                "best_answer_change_time": 0,
            }
        },
    )


class Migration(migrations.Migration):

    dependencies = [("misago_categories", "0006_moderation_queue_roles")]

    operations = [migrations.RunPython(create_default_categories_roles)]



from functools import wraps
from logging import getLogger

from django.contrib.auth import get_user_model, login
from django.http import Http404
from django.shortcuts import redirect, render
from django.urls import reverse
from django.views.decorators.cache import never_cache

from ..core.exceptions import Banned
from ..users.bans import get_user_ban
from ..users.decorators import deny_banned_ips
from ..users.registration import send_welcome_email
from .client import (
    create_login_url,
    get_access_token,
    get_code_grant,
    get_user_data,
)
from .exceptions import (
    OAuth2Error,
    OAuth2UserAccountDeactivatedError,
    OAuth2UserDataValidationError,
)
from .user import get_user_from_data

logger = getLogger("misago.oauth2")

User = get_user_model()


def oauth2_view(f):
    f = deny_banned_ips(f)

    @wraps(f)
    @never_cache
    def wrapped_oauth2_view(request):
        if not request.settings.enable_oauth2_client:
            raise Http404()

        return f(request)

    return wrapped_oauth2_view


@oauth2_view
def oauth2_login(request):
    redirect_to = create_login_url(request)
    return redirect(redirect_to)


@oauth2_view
def oauth2_complete(request):
    try:
        code_grant = get_code_grant(request)
        token = get_access_token(request, code_grant)
        user_data, raw_data = get_user_data(request, token)
        user, created = get_user_from_data(request, user_data, raw_data)

        if not user.is_active:
            raise OAuth2UserAccountDeactivatedError()

        if not user.is_staff:
            if user_ban := get_user_ban(user, request.cache_versions):
                raise Banned(user_ban)
    except OAuth2UserDataValidationError as error:
        logger.exception(
            "OAuth2 Profile Error",
            extra={
                f"error[{error_index}]": str(error_msg)
                for error_index, error_msg in enumerate(error.error_list)
            },
        )
        return render(
            request,
            "misago/errorpages/oauth2_profile.html",
            {
                "error": error,
                "error_list": error.error_list,
            },
            status=400,
        )
    except OAuth2Error as error:
        logger.exception("OAuth2 Error")
        return render(
            request,
            "misago/errorpages/oauth2.html",
            {"error": error},
            status=400,
        )

    if created and request.settings.oauth2_send_welcome_email:
        send_welcome_email(request, user)

    if not user.requires_activation:
        login(request, user)

    return redirect(reverse("misago:index"))



from django.test import TestCase

from ...categories.models import Category
from ..threadtypes.treesmap import TreesMap

THREAD_TYPE = "misago.threads.threadtypes.thread.Thread"


class TreesMapTests(TestCase):
    def test_load_types(self):
        """TreesMap().load_types() loads types modules"""
        trees_map = TreesMap(None)
        types = trees_map.load_types([THREAD_TYPE])

        self.assertEqual(len(types), 1, "expected to load only one thread type")
        self.assertIn("root_category", types, "invalid thread type was loaded")

    def test_load_trees(self):
        """TreesMap().load_trees() loads trees ids"""
        trees_map = TreesMap(None)
        types = trees_map.load_types([THREAD_TYPE])
        trees = trees_map.load_trees(types)

        tree_id = Category.objects.get(special_role="root_category").tree_id

        self.assertEqual(len(trees), 1, "expected to load only one tree")
        self.assertEqual(
            trees[tree_id].root_name, "root_category", "invalid tree was loaded"
        )

    def test_get_roots(self):
        """TreesMap().get_roots() returns roots to trees dict"""
        trees_map = TreesMap(None)
        types = trees_map.load_types([THREAD_TYPE])
        trees = trees_map.load_trees(types)
        roots = trees_map.get_roots(trees)

        tree_id = Category.objects.get(special_role="root_category").tree_id

        self.assertEqual(len(roots), 1, "expected to load only one root")
        self.assertIn("root_category", roots, "invalid root was loaded")
        self.assertEqual(roots["root_category"], tree_id, "invalid tree_id was loaded")

    def test_load(self):
        """TreesMap().load() populates trees map"""
        trees_map = TreesMap([THREAD_TYPE])

        self.assertFalse(
            trees_map.is_loaded, "trees map should be not loaded by default"
        )

        trees_map.load()

        self.assertTrue(
            trees_map.is_loaded, "trees map should be loaded after call to load()"
        )

        self.assertEqual(len(trees_map.types), 1, "expected to load one type")
        self.assertEqual(len(trees_map.trees), 1, "expected to load one tree")
        self.assertEqual(len(trees_map.roots), 1, "expected to load one root")

        tree_id = Category.objects.get(special_role="root_category").tree_id

        self.assertIn(
            "root_category", trees_map.types, "invalid thread type was loaded"
        )
        self.assertEqual(
            trees_map.trees[tree_id].root_name,
            "root_category",
            "invalid tree was loaded",
        )
        self.assertIn("root_category", trees_map.roots, "invalid root was loaded")

    def test_get_type_for_tree_id(self):
        """TreesMap().get_type_for_tree_id() returns type for valid id"""
        trees_map = TreesMap([THREAD_TYPE])
        trees_map.load()

        tree_id = Category.objects.get(special_role="root_category").tree_id
        thread_type = trees_map.get_type_for_tree_id(tree_id)

        self.assertEqual(
            thread_type.root_name,
            "root_category",
            "returned invalid thread type for given tree id",
        )

        try:
            trees_map.get_type_for_tree_id(tree_id + 1000)
            self.fail("invalid tree id should cause KeyError being raised")
        except KeyError as e:
            self.assertIn(
                "tree id has no type defined",
                str(e),
                "invalid exception message as given",
            )

    def test_get_tree_id_for_root(self):
        """TreesMap().get_tree_id_for_root() returns tree id for valid type name"""
        trees_map = TreesMap([THREAD_TYPE])
        trees_map.load()

        in_db_tree_id = Category.objects.get(special_role="root_category").tree_id
        tree_id = trees_map.get_tree_id_for_root("root_category")

        self.assertEqual(
            tree_id, in_db_tree_id, "root name didn't match one in database"
        )

        try:
            trees_map.get_tree_id_for_root("hurr_durr")
            self.fail("invalid root name should cause KeyError being raised")
        except KeyError as e:
            self.assertIn(
                '"hurr_durr" root has no tree defined',
                str(e),
                "invalid exception message as given",
            )



import os

from django.urls import reverse

from .. import test
from ...acl import useracl
from ...acl.objectacl import add_acl_to_obj
from ...categories.models import Category
from ...conftest import get_cache_versions
from ...users.test import AuthenticatedUserTestCase
from ..models import Attachment
from ..serializers import AttachmentSerializer
from ..test import patch_category_acl

TESTFILES_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "testfiles")
TEST_DOCUMENT_PATH = os.path.join(TESTFILES_DIR, "document.pdf")

cache_versions = get_cache_versions()


class EditorApiTestCase(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()

        self.category = Category.objects.get(slug="first-category")


class ThreadPostEditorApiTests(EditorApiTestCase):
    def setUp(self):
        super().setUp()

        self.api_link = reverse("misago:api:thread-editor")

    def test_anonymous_user_request(self):
        """endpoint validates if user is authenticated"""
        self.logout_user()

        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "You need to be signed in to start threads."}
        )

    @patch_category_acl({"can_browse": False})
    def test_category_visibility_validation(self):
        """endpoint omits non-browseable categories"""
        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {
                "detail": (
                    "No categories that allow new threads are available "
                    "to you at the moment."
                )
            },
        )

    @patch_category_acl({"can_start_threads": False})
    def test_category_disallowing_new_threads(self):
        """endpoint omits category disallowing starting threads"""
        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {
                "detail": (
                    "No categories that allow new threads are available "
                    "to you at the moment."
                )
            },
        )

    @patch_category_acl({"can_close_threads": False, "can_start_threads": True})
    def test_category_closed_disallowing_new_threads(self):
        """endpoint omits closed category"""
        self.category.is_closed = True
        self.category.save()

        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {
                "detail": (
                    "No categories that allow new threads are available "
                    "to you at the moment."
                )
            },
        )

    @patch_category_acl({"can_close_threads": True, "can_start_threads": True})
    def test_category_closed_allowing_new_threads(self):
        """endpoint adds closed category that allows new threads"""
        self.category.is_closed = True
        self.category.save()

        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertEqual(
            response_json[0],
            {
                "id": self.category.pk,
                "name": self.category.name,
                "level": 0,
                "post": {"close": True, "hide": False, "pin": 0},
            },
        )

    @patch_category_acl({"can_start_threads": True})
    def test_category_allowing_new_threads(self):
        """endpoint adds category that allows new threads"""
        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertEqual(
            response_json[0],
            {
                "id": self.category.pk,
                "name": self.category.name,
                "level": 0,
                "post": {"close": False, "hide": False, "pin": 0},
            },
        )

    @patch_category_acl({"can_close_threads": True, "can_start_threads": True})
    def test_category_allowing_closing_threads(self):
        """endpoint adds category that allows new closed threads"""
        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertEqual(
            response_json[0],
            {
                "id": self.category.pk,
                "name": self.category.name,
                "level": 0,
                "post": {"close": True, "hide": False, "pin": 0},
            },
        )

    @patch_category_acl({"can_start_threads": True, "can_pin_threads": 1})
    def test_category_allowing_locally_pinned_threads(self):
        """endpoint adds category that allows locally pinned threads"""
        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertEqual(
            response_json[0],
            {
                "id": self.category.pk,
                "name": self.category.name,
                "level": 0,
                "post": {"close": False, "hide": False, "pin": 1},
            },
        )

    @patch_category_acl({"can_start_threads": True, "can_pin_threads": 2})
    def test_category_allowing_globally_pinned_threads(self):
        """endpoint adds category that allows globally pinned threads"""
        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertEqual(
            response_json[0],
            {
                "id": self.category.pk,
                "name": self.category.name,
                "level": 0,
                "post": {"close": False, "hide": False, "pin": 2},
            },
        )

    @patch_category_acl({"can_start_threads": True, "can_hide_threads": 1})
    def test_category_allowing_hidding_threads(self):
        """endpoint adds category that allows hiding threads"""
        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertEqual(
            response_json[0],
            {
                "id": self.category.pk,
                "name": self.category.name,
                "level": 0,
                "post": {"close": 0, "hide": 1, "pin": 0},
            },
        )

    @patch_category_acl({"can_start_threads": True, "can_hide_threads": 2})
    def test_category_allowing_hidding_and_deleting_threads(self):
        """endpoint adds category that allows hiding and deleting threads"""
        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertEqual(
            response_json[0],
            {
                "id": self.category.pk,
                "name": self.category.name,
                "level": 0,
                "post": {"close": False, "hide": 1, "pin": 0},
            },
        )


class ThreadReplyEditorApiTests(EditorApiTestCase):
    def setUp(self):
        super().setUp()

        self.thread = test.post_thread(category=self.category)
        self.api_link = reverse(
            "misago:api:thread-post-editor", kwargs={"thread_pk": self.thread.pk}
        )

    def test_anonymous_user_request(self):
        """endpoint validates if user is authenticated"""
        self.logout_user()

        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "You have to sign in to reply threads."}
        )

    def test_thread_visibility(self):
        """thread's visibility is validated"""
        with patch_category_acl({"can_see": False}):
            response = self.client.get(self.api_link)
            self.assertEqual(response.status_code, 404)

        with patch_category_acl({"can_browse": False}):
            response = self.client.get(self.api_link)
            self.assertEqual(response.status_code, 404)

        with patch_category_acl({"can_see_all_threads": False}):
            response = self.client.get(self.api_link)
            self.assertEqual(response.status_code, 404)

    @patch_category_acl({"can_reply_threads": False})
    def test_no_reply_permission(self):
        """permssion to reply is validated"""
        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "You can't reply to threads in this category."}
        )

    def test_closed_category(self):
        """permssion to reply in closed category is validated"""
        self.category.is_closed = True
        self.category.save()

        with patch_category_acl(
            {"can_reply_threads": True, "can_close_threads": False}
        ):
            response = self.client.get(self.api_link)
            self.assertEqual(response.status_code, 403)
            self.assertEqual(
                response.json(),
                {
                    "detail": (
                        "This category is closed. You can't reply to threads in it."
                    )
                },
            )

        # allow to post in closed category
        with patch_category_acl({"can_reply_threads": True, "can_close_threads": True}):
            response = self.client.get(self.api_link)
            self.assertEqual(response.status_code, 200)

    def test_closed_thread(self):
        """permssion to reply in closed thread is validated"""
        self.thread.is_closed = True
        self.thread.save()

        with patch_category_acl(
            {"can_reply_threads": True, "can_close_threads": False}
        ):
            response = self.client.get(self.api_link)
            self.assertEqual(response.status_code, 403)
            self.assertEqual(
                response.json(),
                {"detail": "You can't reply to closed threads in this category."},
            )

        # allow to post in closed thread
        with patch_category_acl({"can_reply_threads": True, "can_close_threads": True}):
            response = self.client.get(self.api_link)
            self.assertEqual(response.status_code, 200)

    @patch_category_acl({"can_reply_threads": True})
    def test_allow_reply_thread(self):
        """api returns 200 code if thread reply is allowed"""
        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)

    def test_reply_to_visibility(self):
        """api validates replied post visibility"""

        # unapproved reply can't be replied to
        unapproved_reply = test.reply_thread(self.thread, is_unapproved=True)

        with patch_category_acl({"can_reply_threads": True}):
            response = self.client.get(
                "%s?reply=%s" % (self.api_link, unapproved_reply.pk)
            )
            self.assertEqual(response.status_code, 404)

        # hidden reply can't be replied to
        hidden_reply = test.reply_thread(self.thread, is_hidden=True)

        with patch_category_acl({"can_reply_threads": True}):
            response = self.client.get("%s?reply=%s" % (self.api_link, hidden_reply.pk))
            self.assertEqual(response.status_code, 403)
            self.assertEqual(
                response.json(), {"detail": "You can't reply to hidden posts."}
            )

    def test_reply_to_other_thread_post(self):
        """api validates is replied post belongs to same thread"""
        other_thread = test.post_thread(category=self.category)
        reply_to = test.reply_thread(other_thread)

        response = self.client.get("%s?reply=%s" % (self.api_link, reply_to.pk))
        self.assertEqual(response.status_code, 404)

    @patch_category_acl({"can_reply_threads": True})
    def test_reply_to_event(self):
        """events can't be replied to"""
        reply_to = test.reply_thread(self.thread, is_event=True)

        response = self.client.get("%s?reply=%s" % (self.api_link, reply_to.pk))
        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.json(), {"detail": "You can't reply to events."})

    @patch_category_acl({"can_reply_threads": True})
    def test_reply_to(self):
        """api includes replied to post details in response"""
        reply_to = test.reply_thread(self.thread)

        response = self.client.get("%s?reply=%s" % (self.api_link, reply_to.pk))

        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": reply_to.pk,
                "post": reply_to.original,
                "poster": reply_to.poster_name,
            },
        )


class EditReplyEditorApiTests(EditorApiTestCase):
    def setUp(self):
        super().setUp()

        self.thread = test.post_thread(category=self.category)
        self.post = test.reply_thread(self.thread, poster=self.user)

        self.api_link = reverse(
            "misago:api:thread-post-editor",
            kwargs={"thread_pk": self.thread.pk, "pk": self.post.pk},
        )

    def test_anonymous_user_request(self):
        """endpoint validates if user is authenticated"""
        self.logout_user()

        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "You have to sign in to edit posts."}
        )

    def test_thread_visibility(self):
        """thread's visibility is validated"""
        with patch_category_acl({"can_see": False}):
            response = self.client.get(self.api_link)
            self.assertEqual(response.status_code, 404)

        with patch_category_acl({"can_browse": False}):
            response = self.client.get(self.api_link)
            self.assertEqual(response.status_code, 404)

        with patch_category_acl({"can_see_all_threads": False}):
            response = self.client.get(self.api_link)
            self.assertEqual(response.status_code, 404)

    @patch_category_acl({"can_edit_posts": 0})
    def test_no_edit_permission(self):
        """permssion to edit is validated"""
        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "You can't edit posts in this category."}
        )

    def test_closed_category(self):
        """permssion to edit in closed category is validated"""
        self.category.is_closed = True
        self.category.save()

        with patch_category_acl({"can_edit_posts": 1, "can_close_threads": False}):
            response = self.client.get(self.api_link)
            self.assertEqual(response.status_code, 403)
            self.assertEqual(
                response.json(),
                {"detail": "This category is closed. You can't edit posts in it."},
            )

        # allow to edit in closed category
        with patch_category_acl({"can_edit_posts": 1, "can_close_threads": True}):
            response = self.client.get(self.api_link)
            self.assertEqual(response.status_code, 200)

    def test_closed_thread(self):
        """permssion to edit in closed thread is validated"""
        self.thread.is_closed = True
        self.thread.save()

        with patch_category_acl({"can_edit_posts": 1, "can_close_threads": False}):
            response = self.client.get(self.api_link)
            self.assertEqual(response.status_code, 403)
            self.assertEqual(
                response.json(),
                {"detail": "This thread is closed. You can't edit posts in it."},
            )

        # allow to edit in closed thread
        with patch_category_acl({"can_edit_posts": 1, "can_close_threads": True}):
            response = self.client.get(self.api_link)
            self.assertEqual(response.status_code, 200)

    def test_protected_post(self):
        """permssion to edit protected post is validated"""
        self.post.is_protected = True
        self.post.save()

        with patch_category_acl({"can_edit_posts": 1, "can_protect_posts": False}):
            response = self.client.get(self.api_link)
            self.assertEqual(response.status_code, 403)
            self.assertEqual(
                response.json(),
                {"detail": "This post is protected. You can't edit it."},
            )

        # allow to post in closed thread
        with patch_category_acl({"can_edit_posts": 1, "can_protect_posts": True}):
            response = self.client.get(self.api_link)
            self.assertEqual(response.status_code, 200)

    def test_post_visibility(self):
        """edited posts visibility is validated"""
        self.post.is_hidden = True
        self.post.save()

        with patch_category_acl({"can_edit_posts": 1}):
            response = self.client.get(self.api_link)
            self.assertEqual(response.status_code, 403)
            self.assertEqual(
                response.json(), {"detail": "This post is hidden, you can't edit it."}
            )

        # allow hidden edition
        with patch_category_acl({"can_edit_posts": 1, "can_hide_posts": 1}):
            response = self.client.get(self.api_link)
            self.assertEqual(response.status_code, 200)

        # test unapproved post
        self.post.is_unapproved = True
        self.post.is_hidden = False
        self.post.poster = None
        self.post.save()

        with patch_category_acl({"can_edit_posts": 2, "can_approve_content": 0}):
            response = self.client.get(self.api_link)
            self.assertEqual(response.status_code, 404)

        # allow unapproved edition
        with patch_category_acl({"can_edit_posts": 2, "can_approve_content": 1}):
            response = self.client.get(self.api_link)
            self.assertEqual(response.status_code, 200)

    @patch_category_acl({"can_edit_posts": 2})
    def test_post_is_event(self):
        """events can't be edited"""
        self.post.is_event = True
        self.post.save()

        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.json(), {"detail": "Events can't be edited."})

    def test_other_user_post(self):
        """api validates if other user's post can be edited"""
        self.post.poster = None
        self.post.save()

        with patch_category_acl({"can_edit_posts": 1}):
            response = self.client.get(self.api_link)
            self.assertEqual(response.status_code, 403)
            self.assertEqual(
                response.json(),
                {"detail": "You can't edit other users posts in this category."},
            )

        # allow other users post edition
        with patch_category_acl({"can_edit_posts": 2}):
            response = self.client.get(self.api_link)
            self.assertEqual(response.status_code, 200)

    @patch_category_acl({"can_hide_threads": 1, "can_edit_posts": 2})
    def test_edit_first_post_hidden(self):
        """
        endpoint returns valid configuration for editor of hidden thread's first post
        """
        self.thread.is_hidden = True
        self.thread.save()
        self.thread.first_post.is_hidden = True
        self.thread.first_post.save()

        api_link = reverse(
            "misago:api:thread-post-editor",
            kwargs={"thread_pk": self.thread.pk, "pk": self.thread.first_post.pk},
        )

        response = self.client.get(api_link)
        self.assertEqual(response.status_code, 200)

    @patch_category_acl({"can_edit_posts": 1})
    def test_edit(self):
        """endpoint returns valid configuration for editor"""
        with patch_category_acl({"max_attachment_size": 1000}):
            for _ in range(3):
                with open(TEST_DOCUMENT_PATH, "rb") as upload:
                    response = self.client.post(
                        reverse("misago:api:attachment-list"), data={"upload": upload}
                    )
                self.assertEqual(response.status_code, 200)

        attachments = list(Attachment.objects.order_by("id"))

        attachments[0].uploader = None
        attachments[0].save()

        for attachment in attachments[:2]:
            attachment.post = self.post
            attachment.save()

        response = self.client.get(self.api_link)
        user_acl = useracl.get_user_acl(self.user, cache_versions)
        for attachment in attachments:
            add_acl_to_obj(user_acl, attachment)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "id": self.post.pk,
                "api": self.post.get_api_url(),
                "post": self.post.original,
                "can_protect": False,
                "is_protected": self.post.is_protected,
                "poster": self.post.poster_name,
                "attachments": [
                    AttachmentSerializer(
                        attachments[1], context={"user": self.user}
                    ).data,
                    AttachmentSerializer(
                        attachments[0], context={"user": self.user}
                    ).data,
                ],
            },
        )



from unittest.mock import ANY

from ..pipeline import pipeline


def test_markdown_extensions_hook_is_called_by_pipeline(mocker):
    plugin = mocker.Mock()
    mocker.patch("misago.markup.pipeline.hooks.markdown_extensions", [plugin])
    pipeline.extend_markdown(mocker.Mock())
    plugin.asssert_called_once_with(ANY)


def test_parsing_result_processors_hook_is_called_by_pipeline(mocker):
    plugin = mocker.Mock()
    mocker.patch("misago.markup.pipeline.hooks.parsing_result_processors", [plugin])
    pipeline.extend_markdown(mocker.Mock())
    plugin.asssert_called_once_with(ANY, ANY)



from .menuitems import get_footer_menu_items, get_navbar_menu_items


def menus(request):
    return {
        "navbar_menu": get_navbar_menu_items(request.cache_versions),
        "footer_menu": get_footer_menu_items(request.cache_versions),
    }



import pytest

from ..posts import (
    get_fake_hidden_post,
    get_fake_post,
    get_fake_post_content,
    get_fake_unapproved_post,
)
from ..threads import get_fake_thread


@pytest.fixture
def thread(fake, default_category):
    return get_fake_thread(fake, default_category)


def test_fake_post_can_be_created(fake, thread):
    assert get_fake_post(fake, thread)


def test_fake_post_belongs_to_same_category_as_its_thread(fake, thread):
    post = get_fake_post(fake, thread)
    assert post.category == thread.category


def test_fake_post_is_created_with_guest_poster(fake, thread):
    post = get_fake_post(fake, thread)
    assert post.poster is None


def test_fake_post_is_created_with_guest_poster_has_poster_name(fake, thread):
    post = get_fake_post(fake, thread)
    assert post.poster_name


def test_fake_post_is_created_with_specified_poster(fake, thread, user):
    post = get_fake_post(fake, thread, user)
    assert post.poster == user
    assert post.poster_name == user.username


def test_fake_post_is_created_with_valid_checksum(fake, thread):
    post = get_fake_post(fake, thread)
    assert post.is_valid


def test_fake_post_is_created_with_different_content_every_time(fake, thread):
    post_a = get_fake_post(fake, thread)
    post_b = get_fake_post(fake, thread)
    assert post_a.original != post_b.original
    assert post_a.parsed != post_b.parsed


def test_fake_hidden_post_can_be_created(fake, thread):
    post = get_fake_hidden_post(fake, thread)
    assert post.is_hidden


def test_fake_unapproved_post_can_be_created(fake, thread):
    post = get_fake_unapproved_post(fake, thread)
    assert post.is_unapproved


def test_fake_post_content_can_be_created(fake):
    original, parsed = get_fake_post_content(fake)
    assert original
    assert parsed


def test_different_fake_post_content_is_created_every_time(fake):
    original_a, parsed_a = get_fake_post_content(fake)
    original_b, parsed_b = get_fake_post_content(fake)
    assert original_a != original_b
    assert parsed_a != parsed_b
